import inspect
import types
from collections.abc import Sequence
from functools import wraps

import keras
import numpy as np
from keras import ops
from keras.saving import register_keras_serializable
from quantizers import get_fixed_quantizer

from hgq.config import QuantizerConfig
from hgq.constraints import Min, MinMax
from hgq.quantizer.config import KBIConfig
from hgq.quantizer.internal import DefaultBitwidthMapper, FixedPointQuantizerKBI
from hgq.utils.misc import numbers

from .frozen_quantizer import FrozenFixedPointQuantizer
from .utils import parse_string_to_args


@register_keras_serializable(package='hgq/qkeras')
class quantized_bits(QuantizerConfig):
    def __init__(
        self,
        bits: int = 8,
        integer: int = 0,
        symmetric: int = 0,
        keep_negative: bool = True,
        alpha: numbers | str | None = None,
        use_stochastic_rounding: bool = False,
        scale_axis: Sequence[int] | None = None,
        qnoise_factor: float | None = None,
        var_name: None = None,
        use_ste: bool = True,
        use_variables: bool | None = None,
        bc=Min(0),
        br=None,
        ic=MinMax(-12, 12),
        ir=None,
    ):
        """**Configuration** for the creating a qkeras-like quantized_bits quantizer. Can be called to quantize arrays, but does not support training. Pass to the constructor of Quantizer to create a quantizer, or call get_quantizer() to get the internal quantizer (missing certain features realized in the wrapper).

        Parameters
        ----------
        bits : int, optional
            The bit width of the quantized values excluding the sign bit, by default 8
        integer : int, optional
            The number of integer bits excluding the sign bit, by default 0
        symmetric : int, optional
            Whether to use symmetric quantization, by default False
        keep_negative : bool, optional
            Whether to keep negative values, by default True
        alpha : numbers | str | None, optional
            The scaling factor for the quantization. If a number, use the number as the fixed scaling factor. If 'trainable_po2', make number of integer bits trainable. If None, skip scaling, by default 'trainable_po2'
        use_stochastic_rounding : bool, optional
            Whether to use stochastic rounding, by default False
        scale_axis : Sequence[int] | None, optional
            The axes to apply heterogenous quantization, only used if alpha is 'trainable_po2', by default None
        qnoise_factor : float, optional
            The fraction of quantization strength. If None, treat as full quantization noise, by default None
        var_name : None, optional
            Not supported. Must be None, by default None
        use_ste : bool, optional
            Not supported. Must be True, by default True
        use_variables : bool | None, optional
            Whether to use variables for the quantization parameters. Must be True or None if alpha is 'trainable_po2', by default None
        bc : Constraint, optional
           The constraint for the bitwidth. Only used if bitwidth is trainable, by default Min(0)
        br : Regularizer, optional
            The regularizer for the bitwidth. Only used if bitwidth is trainable, by default None
        ic : Constraint, optional
            The constraint for the integer width. Only used if integer width is trainable, by default MinMax(-12, 12)
        ir : Regularizer, optional
            The regularizer for the integer width. Only used if integer width is trainable, by default None
        """

        assert bits > 0, 'bits must be greater than 0'
        if isinstance(alpha, numbers):
            assert alpha > 0, 'alpha must be greater than 0 if it is a number'
        if isinstance(alpha, str):
            assert alpha == 'trainable_po2', "alpha must be 'trainable_po2' if it is a string"
        self.trainable = alpha == 'trainable_po2'

        if use_variables is None:
            use_variables = self.trainable
        if self.trainable:
            assert use_variables, "use_variables must be True if alpha is set to 'trainable_po2'"
        self.use_variables = use_variables

        assert var_name is None, 'var_name is not supported'
        assert use_ste, 'only use_ste=True is supported'

        scaler = alpha if isinstance(alpha, numbers) else None
        heterogeneous_axis = ()
        if scale_axis is not None and self.trainable:
            heterogeneous_axis = tuple(scale_axis)
        overflow_mode = 'SAT_SYM' if symmetric else 'SAT'
        round_mode = 'S_RND_CONV' if use_stochastic_rounding else 'RND_CONV'

        super().__init__(
            'kbi',
            'datalane',
            k0=keep_negative,
            b0=bits,
            i0=integer,
            round_mode=round_mode,
            overflow_mode=overflow_mode,
            bc=bc,
            ic=ic,
            br=br,
            ir=ir,
            i_decay_speed=-1,
            bw_mapper=DefaultBitwidthMapper(heterogeneous_axis=heterogeneous_axis),
            scaler=scaler,
            qnoise_factor=qnoise_factor,
            trainable=self.trainable,
        )
        self.config: KBIConfig
        self.stateless_quantizer = get_fixed_quantizer(round_mode, overflow_mode)

    @property
    def bits(self):
        b = self.config['b0']
        assert isinstance(b, int)
        return b

    @property
    def integers(self):
        i = self.config['i0']
        assert isinstance(i, int)
        return i

    @property
    def symmetric(self):
        return 'SYM' in self.config['overflow_mode'].upper()

    @property
    def keep_negative(self):
        return bool(self.config['k0'])

    def get_quantizer(self):
        if not self.use_variables:
            return FrozenFixedPointQuantizer(**self.config)  # type: ignore

        quantizer = FixedPointQuantizerKBI(**self.config)
        if not self.trainable:
            return quantizer

        # Make the bits non-trainable by default to match qkeras behavior
        original_setattr_hook = quantizer._setattr_hook

        @wraps(original_setattr_hook)
        def _setattr_hook(self, name, value):
            if name == '_b':
                value.trainable = False
            return original_setattr_hook(name, value)

        quantizer._setattr_hook = types.MethodType(_setattr_hook, quantizer)

        return quantizer

    def __str__(self):
        k, b, i = self.config['k0'], self.config['b0'], self.config['i0']
        symmetric = 'SYM' in self.config['overflow_mode'].upper()
        alpha = self.scaler
        if alpha is None:
            alpha = 'trainable_po2' if self.trainable else 1
        stochastic = 'S_RND' in self.config['round_mode']
        cls_str = self.__class__.__name__
        return (
            f'{cls_str}({b}, {i}, symmetric={symmetric}, keep_negative={k}, alpha={alpha}, use_stochastic_rounding={stochastic})'
        )

    def __repr__(self):
        return str(self)

    def __call__(self, inputs, training=None):
        assert training is None, (
            'quantized_bits in HGQ2 is merely a configuration. It does not support training. Call get_quantizer() to get the actual quantizer.'
        )
        return_np = False
        if not ops.is_tensor(inputs):
            return_np = True
            inputs = np.asarray(inputs)
            if inputs.dtype not in (np.float16, np.float32):
                inputs = inputs.astype(np.float32)
        k, b, i = self.keep_negative, self.bits, self.integers

        with keras.device('cpu:0'):
            if self.scaler is not None:
                inputs /= self.scaler
            outputs = self.stateless_quantizer(inputs, k, b, i, False, None)
            if self.scaler is not None:
                outputs *= self.scaler
            if self.qnoise_factor != 1.0:
                outputs = inputs + self.qnoise_factor * (outputs - inputs)  # type: ignore
        if return_np:
            outputs = np.asarray(outputs)
        return outputs

    @property
    def max(self):
        b, i = self.bits, self.integers
        f = b - i
        return 2.0**i - 2.0**-f

    @property
    def min(self):
        if not self.keep_negative:
            return 0.0
        if self.symmetric:
            return -self.max
        else:
            return -(2.0**self.integers)

    def get_config(self):  # type: ignore
        config = dict(
            bits=self.bits,
            integer=self.integers,
            symmetric=self.symmetric,
            keep_negative=self.keep_negative,
            alpha=self.scaler if not self.trainable else 'trainable_po2',
            use_stochastic_rounding='S_RND' in self.config['round_mode'],
            scale_axis=self.config['bw_mapper'].heterogeneous_axis,  # type: ignore
            qnoise_factor=self.qnoise_factor,
            var_name=None,
            use_ste=True,
            use_variables=self.trainable,
            bc=self.config['bc'],
            br=self.config['br'],
            ic=self.config['ic'],
            ir=self.config['ir'],
        )
        return config

    @classmethod
    def from_string(cls, string: str):
        """Create a quantized_bits configuration from a string representation."""
        if not string.startswith('quantized_bits(') or not string.endswith(')'):
            raise ValueError(f'Invalid string representation for quantized_bits: {string}')
        args, kwargs = parse_string_to_args(string.split('(', 1)[1][:-1])
        signature = inspect.signature(cls.__init__)
        # remove self
        signature = signature.replace(parameters=list(signature.parameters.values())[1:])
        bound = signature.bind(*args, **kwargs)
        bound.apply_defaults()
        return cls(**bound.arguments)


class quantized_relu(quantized_bits):
    def __init__(
        self,
        bits: int = 8,
        integer: int = 0,
        alpha: numbers | str | None = 1.0,
        use_stochastic_rounding: bool = False,
        scale_axis: Sequence[int] | None = None,
        qnoise_factor: float | None = None,
        var_name: None = None,
        use_ste: bool = True,
        use_variables: bool | None = None,
        bc=Min(0),
        br=None,
        ic=MinMax(-12, 12),
        ir=None,
    ):
        """Grammatical sugar for quantized_bits with keep_negative=False"""
        ...

    def __new__(cls, *args, **kwargs):
        return quantized_bits(*args, **kwargs, keep_negative=False)



from .linformer import QLinformerAttention
from .mha import QMultiHeadAttention
from .salt import QSALTAttention

__all__ = [
    'QLinformerAttention',
    'QMultiHeadAttention',
    'QSALTAttention',
]



import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

import keras
import numpy as np
from keras.callbacks import Callback


class ParetoFront(Callback):
    def __init__(
        self,
        path: str | Path,
        metrics: list[str],
        sides: list[int],
        fname_format: str | None = None,
        enable_if: Callable[[dict[str, Any]], bool] | None = None,
    ):
        if fname_format is None:
            fname_format = 'epoch={epoch}'
            for metric in metrics:
                fname_format += f'_{metric}={{{metric}}}'
            fname_format += '.keras'
        else:
            fname_format = fname_format.strip()
        self.path = Path(path)
        self.paths = []
        self.record = []
        self.metrics = metrics
        self.sides = np.array(sides)
        self.enable_if = enable_if
        self.fname_format = fname_format

        _fname_format = fname_format.lower()
        if _fname_format.endswith('.weights.h5') or _fname_format.endswith('.weights.json'):
            self._save_weights = True
        else:
            self._save_weights = False

    def on_train_begin(self, logs=None):
        os.makedirs(self.path, exist_ok=True)

    def on_epoch_end(self, epoch, logs=None):
        assert isinstance(self.model, keras.models.Model)
        assert isinstance(logs, dict)

        logs = logs.copy()
        logs['epoch'] = epoch

        if self.enable_if is not None and not self.enable_if(logs):
            return
        new_metrics = np.array([logs[metric_name] for metric_name in self.metrics])
        _rm_idx = []
        for i, old_metrics in enumerate(self.record):
            _old_metrics = self.sides * old_metrics
            _new_metrics = self.sides * new_metrics
            if np.all(_new_metrics <= _old_metrics):
                return
            if np.all(_new_metrics > _old_metrics):
                _rm_idx.append(i)
        for i in _rm_idx[::-1]:
            self.record.pop(i)
            p = self.paths.pop(i)
            os.remove(p)

        path = self.path / self.fname_format.format(**logs)
        self.record.append(new_metrics)
        self.paths.append(path)
        if self._save_weights:
            self.model.save_weights(path)
        else:
            self.model.save(path)



from math import ceil
from random import randint as rnd_uniform
from random import shuffle as rnd_shuffle
from warnings import warn

import keras
from keras import ops
from keras.utils import PyDataset


class Dataset(PyDataset):
    def __init__(self, x_set, y_set=None, batch_size=None, device: str = 'cpu:0', drop_last=False, shuffle=False, **kwargs):
        super().__init__(**kwargs)

        self.shuffle = shuffle

        n_inp = 1
        if isinstance(x_set, (list, tuple)):
            n_inp = len(x_set)
            if n_inp > 10:
                warn(f'You have {n_inp} input tensors. Are you sure this is correct?')

        n_out = 1 if y_set is not None else None
        if isinstance(y_set, (list, tuple)):
            n_out = len(y_set)
            if n_out > 10:
                warn(f'You have {n_out} output tensors. Are you sure this is correct?')

        self.n_inp = n_inp
        self.n_out = n_out

        to_tensor = ops.convert_to_tensor if device != 'none' else lambda x: x
        self.device = device

        device = 'cpu:0' if device == 'none' else device

        with keras.device(device):
            if self.n_inp == 1:
                self.x = to_tensor(x_set)
                self.data_len = len(x_set)
            else:
                self.x = tuple(to_tensor(x) for x in x_set)
                self.data_len = len(x_set[0])
                assert all(len(x) == self.data_len for x in x_set), 'All input tensors must have the same length'
            if y_set is not None:
                if self.n_out == 1:
                    self.y = to_tensor(y_set)
                    assert len(y_set) == self.data_len, 'Output tensor must have the same length as input tensor'
                else:
                    self.y = tuple(to_tensor(y) for y in y_set)
                    assert all(len(y) == self.data_len for y in y_set), (
                        'All output tensors must have the same length as input tensor'
                    )
            else:
                self.y = None

        self.drop_last = drop_last
        self.batch(batch_size)

    def __len__(self):
        assert self.batch_size is not None, 'batch_size must be set'
        if self.drop_last:
            return self.data_len // self.batch_size
        return ceil(self.data_len / self.batch_size)

    def batch(self, batch_size):
        self.batch_size = batch_size
        if self.batch_size is not None:
            self._reminder = self.data_len % self.batch_size
            self._init_shift = rnd_uniform(0, self._reminder)
        else:
            self._reminder = 0
            self._init_shift = 0

        if self.shuffle and self.batch_size is not None:
            self._index_map = list(range(len(self)))
            rnd_shuffle(self._index_map)
        else:
            self._index_map = None

    def __getitem__(self, idx: int):
        assert type(self.batch_size) is int, 'batch_size must be set before getting items'

        if idx < 0:
            idx = idx % len(self)
        assert idx < len(self), f'Index out of range: {idx} >= {len(self)}'

        if self._index_map is not None:
            idx = self._index_map[idx]

        low = idx * self.batch_size
        high = min(low + self.batch_size, self.data_len)

        if self.drop_last:
            low += self._init_shift
            high += self._init_shift

        if self.n_inp == 1:
            batch_x = self.x[low:high]  # type: ignore
        else:
            batch_x = tuple(x[low:high] for x in self.x)  # type: ignore
        if self.y is not None:
            if self.n_out == 1:
                batch_y = self.y[low:high]  # type: ignore
            else:
                batch_y = tuple(y[low:high] for y in self.y)  # type: ignore
        else:
            batch_y = float('nan')
        return batch_x, batch_y

    def on_epoch_end(self):
        if self._index_map is not None:
            rnd_shuffle(self._index_map)

    def on_epoch_begin(self):
        if self.drop_last:
            self._init_shift = rnd_uniform(0, self._reminder)



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

import keras
from da4ml.converter.plugin import DAISTracerPluginBase, _flatten_arr
from da4ml.trace import FixedVariableArray, comb_trace
from keras import KerasTensor, Operation

from .layers import _registry


@dataclass
class OpObj:
    operation: Operation
    args: list
    kwargs: dict
    produces: tuple[KerasTensor, ...]
    requires: tuple[KerasTensor, ...]


def parse_model(model: keras.Model):
    if isinstance(model, keras.Sequential):
        model = model._functional
    operators: dict[int, list[OpObj]] = {}
    for depth, nodes in model._nodes_by_depth.items():
        _oprs = []
        for node in nodes:
            assert isinstance(node.operation, keras.Operation)
            opr = OpObj(
                operation=node.operation,
                args=node.arguments.args,
                kwargs=node.arguments.kwargs,
                produces=node.outputs,
                requires=node.arguments.keras_tensors,
            )
            _oprs.append(opr)
        operators[depth] = _oprs
    return [operators[i] for i in range(max(operators.keys()), -1, -1)]


def replace_tensors(tensor_map: dict[KerasTensor, FixedVariableArray], obj: Any) -> Any:
    if isinstance(obj, KerasTensor):
        return tensor_map[obj]
    if isinstance(obj, list):
        return [replace_tensors(tensor_map, o) for o in obj]
    if isinstance(obj, tuple):
        return tuple(replace_tensors(tensor_map, o) for o in obj)
    if isinstance(obj, dict):
        return {k: replace_tensors(tensor_map, v) for k, v in obj.items()}
    return obj


def _apply_nn(
    model: keras.Model,
    inputs: FixedVariableArray | Sequence[FixedVariableArray],
    verbose: bool = False,
    dump: bool = False,
    n_nested: int = 0,
) -> tuple[FixedVariableArray, ...] | dict[str, FixedVariableArray]:
    """
    Apply a keras model to a fixed variable array or a sequence of fixed variable arrays.

    Parameters
    ----------
    model : keras.Model
        The keras model to apply.
    inputs : FixedVariableArray or Sequence[FixedVariableArray]
        The input fixed variable array or sequence of fixed variable arrays.

    Returns
    -------
    tuple of FixedVariableArray
        A tuple containing the output(s) of the model as FixedVariableArray.
    """
    if isinstance(inputs, FixedVariableArray):
        inputs = (inputs,)

    assert len(model.inputs) == len(inputs), f'Model has {len(model.inputs)} inputs, got {len(inputs)}'
    tensor_map = {keras_tensor: da_tensor for keras_tensor, da_tensor in zip(model.inputs, inputs)}

    _inputs = _flatten_arr(inputs)

    if verbose and n_nested:
        print(' -> enter:')

    for ops in parse_model(model):
        for op in ops:
            assert all(t in tensor_map for t in op.requires)
            args = replace_tensors(tensor_map, op.args)
            kwargs: dict[str, Any] = replace_tensors(tensor_map, op.kwargs)
            if op.operation.__class__ is keras.layers.InputLayer:
                continue

            if verbose:
                indent = '    ' * n_nested
                print(f'{indent}{op.operation.name} ({op.operation.__class__.__name__})', end='')

            if isinstance(op.operation, keras.Model):
                sub_model = op.operation._functional if isinstance(op.operation, keras.Sequential) else op.operation
                outputs: tuple[FixedVariableArray, ...] = _apply_nn(
                    sub_model,
                    args,
                    verbose=verbose,
                    dump=False,
                    n_nested=n_nested + 1,
                )  # type: ignore
            else:
                mirror_op = _registry[op.operation.__class__](op.operation)
                outputs = mirror_op(*args, **kwargs)
            if verbose:
                comb = comb_trace(_inputs, _flatten_arr(outputs))
                print(f' cumcost: {comb.cost}, latency: {comb.latency[1]}')

            for keras_tensor, da_tensor in zip(op.produces, outputs):
                tensor_map[keras_tensor] = da_tensor

    if verbose and n_nested:
        indent = '    ' * (n_nested - 1)
        print(f'{indent}<- exit', end='')

    if not dump:
        return tuple(tensor_map[keras_tensor] for keras_tensor in model.outputs)
    else:
        return {k.name: v for k, v in tensor_map.items()}


class HGQ2DAISTracer(DAISTracerPluginBase):
    model: keras.Model

    def apply_model(self, verbose: bool, inputs: tuple[FixedVariableArray, ...]):
        dump: dict[str, FixedVariableArray] = _apply_nn(
            self.model,
            inputs,
            verbose=verbose,
            dump=True,
        )  # type: ignore
        output_names = [out.name for out in self.model.outputs]
        return dump, output_names

    def get_input_shapes(self) -> list[tuple[int, ...]] | None:
        shapes = [inp.shape[1:] for inp in self.model.inputs]
        if any(None in shape or -1 in shape for shape in shapes):
            return None
        return shapes



from ..dataset import Dataset
from .beta_pid import BetaPID
from .beta_scheduler import BetaScheduler, PieceWiseSchedule
from .early_stopping_ebops import EarlyStoppingWithEbopsThres
from .ebops import FreeEBOPs
from .pareto import ParetoFront
from .pbar import PBar

__all__ = [
    'BetaPID',
    'BetaScheduler',
    'PieceWiseSchedule',
    'Dataset',
    'EarlyStoppingWithEbopsThres',
    'FreeEBOPs',
    'PBar',
    'ParetoFront',
]



from math import log2, prod

import keras
import numpy as np
import pytest
from keras import layers

from hgq.layers import QLayerBase
from hgq.layers.pooling import (
    QAveragePooling1D,
    QAveragePooling2D,
    QGlobalAveragePooling1D,
    QGlobalAveragePooling2D,
    QGlobalMaxPooling1D,
    QGlobalMaxPooling2D,
    QMaxPooling1D,
    QMaxPooling2D,
)

from .base import LayerTestBase


class PoolingTestBase(LayerTestBase):
    layer_cls = QLayerBase
    keras_layer_cls = layers.Layer
    dim = -1

    @pytest.fixture
    def input_shapes(self):
        shape = (5, 8, 5)[: self.dim + 1]
        return shape

    @pytest.fixture(params=['valid', 'same'])
    def padding(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, pool_size, strides, padding):
        return {
            'pool_size': pool_size,
            'strides': strides,
            'padding': padding,
        }

    def test_hls4ml_conversion(
        self, model: keras.Model, input_data: np.ndarray, temp_directory: str, use_parallel_io: bool, q_type: str
    ):
        layer = model.layers[-1]
        if not use_parallel_io and getattr(layer, 'padding', None) == 'same':
            pytest.skip("hls4ml does not support 'same' padding without parallel IO")
        if self.dim == 2 and isinstance(layer, (QAveragePooling2D, QMaxPooling2D)) and layer.strides != layer.pool_size:
            pytest.skip('Known bug in hls4ml vivado/vitis + io stream with 2d pooling with non-trivial strides')
        super().test_hls4ml_conversion(
            model=model,
            input_data=input_data,
            temp_directory=temp_directory,
            use_parallel_io=use_parallel_io,
            q_type=q_type,
        )

    def test_da4ml_conversion(self, model: keras.Model, input_data, overflow_mode: str, q_type: str):
        layer = model.layers[-1]

        if isinstance(layer, (QAveragePooling2D, QAveragePooling1D)):
            if getattr(layer, 'padding', None) == 'same':
                pytest.skip('non-pow-2 pool size at boundary')
            if log2(prod(layer.pool_size)) % 1 != 0:
                pytest.skip('non-pow-2 pool size')

        super().test_da4ml_conversion(
            model=model,
            input_data=input_data,
            overflow_mode=overflow_mode,
            q_type=q_type,
        )


class GlobalPoolingTestBase(PoolingTestBase):
    @pytest.fixture
    def input_shapes(self):
        shape = (4, 8, 5)[: self.dim + 1]
        return shape

    @pytest.fixture
    def layer_kwargs(self):
        return {'data_format': 'channels_last'}


class TestQMaxPooling1D(PoolingTestBase):
    layer_cls = QMaxPooling1D
    keras_layer_cls = layers.MaxPooling1D
    dim = 1

    @pytest.fixture(params=[2, 4])
    def pool_size(self, request):
        return request.param

    @pytest.fixture(params=[3])
    def strides(self, request):
        return request.param


class TestQAveragePooling1D(PoolingTestBase):
    layer_cls = QAveragePooling1D
    keras_layer_cls = layers.AveragePooling1D
    dim = 1

    @pytest.fixture(params=[2, 4])
    def pool_size(self, request):
        return request.param

    @pytest.fixture(params=[3])
    def strides(self, request):
        return request.param

    def assert_equal(self, keras_output, hls_output):
        return np.testing.assert_allclose(keras_output, hls_output, rtol=0, atol=5e-6)


class TestQMaxPooling2D(PoolingTestBase):
    layer_cls = QMaxPooling2D
    keras_layer_cls = layers.MaxPooling2D
    dim = 2

    @pytest.fixture(params=[(2, 4)])
    def pool_size(self, request):
        return request.param

    @pytest.fixture(params=[(3, 2), (2, 4)])
    def strides(self, request):
        return request.param

    @pytest.fixture(params=['channels_last'])
    def layer_kwargs(self, request, pool_size, strides, padding):
        return {
            'pool_size': pool_size,
            'strides': strides,
            'padding': padding,
            'data_format': request.param,
        }


class TestQAveragePooling2D(PoolingTestBase):
    layer_cls = QAveragePooling2D
    keras_layer_cls = layers.AveragePooling2D
    dim = 2

    @pytest.fixture(params=[(2, 4)])
    def pool_size(self, request):
        return request.param

    @pytest.fixture(params=[(3, 2), (2, 4)])
    def strides(self, request):
        return request.param

    @pytest.fixture(params=['channels_last'])
    def layer_kwargs(self, request, pool_size, strides, padding):
        return {
            'pool_size': pool_size,
            'strides': strides,
            'padding': padding,
            'data_format': request.param,
        }

    def assert_equal(self, keras_output, hls_output):
        return np.testing.assert_allclose(keras_output, hls_output, rtol=0, atol=5e-6)


class TestQGlobalMaxPooling1D(GlobalPoolingTestBase):
    layer_cls = QGlobalMaxPooling1D
    keras_layer_cls = layers.GlobalMaxPooling1D
    dim = 1


class TestQGlobalMaxPooling2D(GlobalPoolingTestBase):
    layer_cls = QGlobalMaxPooling2D
    keras_layer_cls = layers.GlobalMaxPooling2D
    dim = 2


class TestQGlobalAveragePooling1D(GlobalPoolingTestBase):
    layer_cls = QGlobalAveragePooling1D
    keras_layer_cls = layers.GlobalAveragePooling1D
    dim = 1

    def assert_equal(self, keras_output, hls_output):
        return np.testing.assert_allclose(keras_output, hls_output, rtol=0, atol=5e-6)


class TestQGlobalAveragePooling2D(GlobalPoolingTestBase):
    layer_cls = QGlobalAveragePooling2D
    keras_layer_cls = layers.GlobalAveragePooling2D
    dim = 2

    def assert_equal(self, keras_output, hls_output):
        return np.testing.assert_allclose(keras_output, hls_output, rtol=0, atol=5e-6)



from keras import backend, initializers, ops
from keras.constraints import Constraint
from keras.regularizers import Regularizer

from hgq.constraints import MinMax
from hgq.quantizer.internal import FixedPointQuantizerKBI, numbers
from hgq.utils.misc import gather_vars_to_kwargs


class FrozenFixedPointQuantizer(FixedPointQuantizerKBI):
    """Abstract base class for all fixed-point quantizers."""

    def __init__(
        self,
        k0: numbers | bool,
        b0: numbers,
        i0: numbers,
        round_mode: str,
        overflow_mode: str,
        bc: Constraint | None = MinMax(0, 12),
        ic: Constraint | None = None,
        br: Regularizer | None = None,
        ir: Regularizer | None = None,
        **kwargs,
    ):
        trainable = kwargs.pop('trainable', False)
        assert not trainable, 'FrozenFixedPointQuantizer does not support trainable=True'

        kwargs = gather_vars_to_kwargs('self')
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.building = True
        super().build(input_shape)
        del self.building

    def add_weight(  # type: ignore
        self,
        shape=None,
        initializer=None,
        dtype=None,
        trainable=True,
        autocast=True,
        regularizer=None,
        constraint=None,
        aggregation='mean',
        name=None,
    ):
        """Override add_weight to make it adds a constant weight."""
        if dtype is not None:
            dtype = backend.standardize_dtype(dtype)
        else:
            dtype = self.variable_dtype
        initializer = initializers.get(initializer)
        assert initializer is not None, f'Could not interpret initializer: {initializer}'
        value = initializer(shape, dtype=dtype)
        if ops.size(value) == 1:
            value = ops.squeeze(value)
        return value

    def get_config(self):
        config = super().get_config()
        config.pop('trainable')
        return config



import numpy as np
from da4ml.trace import FixedVariableArray

from hgq.layers import QBatchNormalization

from ._base import ReplayOperationBase


class ReplayQBatchNormalization(ReplayOperationBase):
    handles = (QBatchNormalization,)

    def call(self, inputs: FixedVariableArray, mask=None) -> FixedVariableArray:
        layer: QBatchNormalization = self.op
        scale, bias = map(np.array, layer.qscaler_and_qoffset)
        shape = layer._shape[1:]
        return inputs * scale.reshape(shape) + bias.reshape(shape)



import keras
import numpy as np
import pytest

from hgq.config import LayerConfigScope, QuantizerConfigScope
from hgq.layers import QConvT1D, QConvT2D, QDenseT

from .base import CtxGlue, LayerTestBase


class TableTestBase(LayerTestBase):
    @pytest.fixture(params=[True])
    def use_parallel_io(self, request) -> bool:
        return request.param

    @pytest.fixture(params=[True, False])
    def batch_norm(self, request) -> bool:
        return request.param

    @pytest.fixture
    def ctx_scope(self, q_type: str, overflow_mode: str, round_mode: str):
        scope_t = QuantizerConfigScope(place='table', homogeneous_axis=(0,), overflow_mode='SAT')
        scope_w = QuantizerConfigScope(
            default_q_type=q_type,
            heterogeneous_axis=None,
            homogeneous_axis=(),
            overflow_mode=overflow_mode,
            round_mode=round_mode,
            br=None,
            ir=None,
            fr=None,
        )
        scope_a = QuantizerConfigScope(
            default_q_type=q_type, place='datalane', overflow_mode=overflow_mode, round_mode=round_mode, homogeneous_axis=(0,)
        )
        scope_l = LayerConfigScope(beta0=0.0, enable_ebops=True)
        return CtxGlue(scope_w, scope_a, scope_t, scope_l)


class TestDenseT(TableTestBase):
    layer_cls = QDenseT

    @pytest.fixture(params=[8, 12])  # Test different output sizes
    def n_out(self, request):
        return request.param

    @pytest.fixture(params=[None])  # Test with and without activation
    def activation(self, request):
        return request.param

    @pytest.fixture(params=[(8, 8), (12,)])
    def input_shapes(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, n_out, activation, batch_norm):
        return {'n_out': n_out, 'activation': activation, 'batch_norm': batch_norm}


class TestConvT1D(TableTestBase):
    layer_cls = QConvT1D

    @pytest.fixture(params=[8, 12])
    def ch_out(self, request):
        return request.param

    @pytest.fixture(
        params=[
            {'kernel_size': 1, 'strides': 2, 'parallelization_factor': -1},
            {'kernel_size': 3, 'strides': 1, 'parallelization_factor': -1},
            {'kernel_size': 2, 'strides': 2, 'parallelization_factor': 1},
        ]
    )
    def conv_params(self, request):
        return request.param

    @pytest.fixture()
    def input_shapes(self):
        return (6, 2)

    @pytest.fixture(params=['valid', 'same'])
    def padding(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, ch_out, conv_params, batch_norm):
        return {'filters': ch_out, 'batch_norm': batch_norm, **conv_params}


class TestConvT2D(TableTestBase):
    layer_cls = QConvT2D

    @pytest.fixture(params=[8, 12])
    def ch_out(self, request):
        return request.param

    @pytest.fixture(
        params=[
            {'kernel_size': (1, 1), 'strides': (3, 5), 'parallelization_factor': -1},
            {'kernel_size': (3, 3), 'strides': (1, 1), 'parallelization_factor': -1},
            {'kernel_size': (3, 2), 'strides': (1, 3), 'parallelization_factor': 1},
            {'kernel_size': (2, 3), 'strides': (4, 2), 'parallelization_factor': 1},
        ]
    )
    def conv_params(self, request):
        return request.param

    @pytest.fixture()
    def input_shapes(self):
        return (6, 6, 2)

    @pytest.fixture(params=['valid', 'same'])
    def padding(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, ch_out, conv_params, batch_norm):
        return {'filters': ch_out, 'batch_norm': batch_norm, **conv_params}

    def test_training(self, model: keras.Model, input_data: np.ndarray, ch_out: int):
        if keras.backend.backend() == 'torch' and ch_out == 1:
            pytest.skip('Torch runtime error for unknown reason when ch_out is 1.')
        return super().test_training(model, input_data, 'SAT')



from math import prod

import keras
import numpy as np
from da4ml.trace import FixedVariableArray
from keras.src.layers.pooling.base_global_pooling import BaseGlobalPooling
from keras.src.layers.pooling.base_pooling import BasePooling

import hgq

from ._base import ReplayOperationBase
from .conv import symbolic_extract_patches


class ReplayPool(ReplayOperationBase):
    handles = (
        hgq.layers.QAvgPool1D,
        hgq.layers.QAvgPool2D,
        hgq.layers.QAvgPool3D,
        hgq.layers.QMaxPool1D,
        hgq.layers.QMaxPool2D,
        hgq.layers.QMaxPool3D,
        hgq.layers.QGlobalAveragePooling1D,
        hgq.layers.QGlobalMaxPooling1D,
        hgq.layers.QGlobalAveragePooling2D,
        hgq.layers.QGlobalMaxPooling2D,
        hgq.layers.QGlobalAveragePooling3D,
        hgq.layers.QGlobalMaxPooling3D,
        keras.layers.AveragePooling1D,
        keras.layers.AveragePooling2D,
        keras.layers.AveragePooling3D,
        keras.layers.MaxPooling1D,
        keras.layers.MaxPooling2D,
        keras.layers.MaxPooling3D,
        keras.layers.GlobalAveragePooling1D,
        keras.layers.GlobalMaxPooling1D,
        keras.layers.GlobalAveragePooling2D,
        keras.layers.GlobalMaxPooling2D,
        keras.layers.GlobalAveragePooling3D,
        keras.layers.GlobalMaxPooling3D,
    )

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        cname = self.op.__class__.__name__
        if 'Max' in cname:
            op = 'max'
        else:
            assert 'Average' in cname, f'Unsupported global pooling layer: {cname}'
            op = 'avg'

        data_format = self.op.data_format
        if data_format == 'channels_first':
            inputs = np.moveaxis(inputs, 1, -1)  # type: ignore

        if isinstance(self.op, BaseGlobalPooling):
            pool_dim = self.op.input_spec.ndim - 2  # type: ignore
            axis = tuple(range(pool_dim))
            keepdims = self.op.keepdims

            if op == 'max':
                out = np.amax(inputs, axis=axis, keepdims=keepdims)  # type: ignore
            elif op == 'avg':
                pool_size = prod(inputs.shape[:-1])
                out = np.sum(inputs, axis=axis, keepdims=keepdims) / pool_size  # type: ignore
        else:
            assert isinstance(self.op, BasePooling), f'Unknown pooling layer: {type(self.op)}'
            pool_size = self.op.pool_size
            strides = self.op.strides
            padding = self.op.padding
            pool_dim = len(pool_size)
            ch = inputs.shape[-1]
            x = symbolic_extract_patches(
                inputs,
                pool_size,
                strides,
                dilation_rate=1,
                padding=padding,
                data_format='channels_last',
            )
            x = x.reshape(x.shape[:-1] + (-1, ch))

            if padding == 'same':
                mask = symbolic_extract_patches(
                    np.ones(inputs.shape, dtype=np.int32),
                    pool_size,
                    strides,
                    dilation_rate=1,
                    padding=padding,
                    data_format='channels_last',
                ).reshape(x.shape)
            elif padding == 'valid':
                mask = np.ones(x.shape, dtype=np.int32)
            else:
                raise ValueError(f'Unknown padding type: {padding}')

            if op == 'max':
                _vars = np.where(mask, x._vars, -(65535**2))
                x = FixedVariableArray(_vars, x.solver_options)
                out = np.max(x, axis=-2)  # type: ignore
            elif op == 'avg':
                out = np.sum(x, axis=-2) / np.sum(mask, axis=-2)  # type: ignore
            else:
                raise ValueError(f'Unknown pooling operation: {op}')

        if data_format == 'channels_first':
            out = np.moveaxis(out, -1, 1)  # type: ignore

        return out  # type: ignore



import numpy as np
from da4ml.trace import FixedVariableArray
from da4ml.trace.ops import einsum

from hgq.layers.attn import (
    QLinformerAttention,
    QMultiHeadAttention,
    QSALTAttention,
)

from ._base import ReplayOperationBase, mirror_quantizer, to_np_arr
from .activation import ReplayQSoftmax
from .conv import ReplayQConv
from .dense import ReplayQDense


def _fused_qkv_proj(query: FixedVariableArray, key: FixedVariableArray, value: FixedVariableArray, op: QMultiHeadAttention):
    query_qk, query_qb = op.query_dense.qkernel, op.query_dense.qbias
    query_qk = to_np_arr(query_qk)
    query_qb = to_np_arr(query_qb) if query_qb is not None else np.zeros_like(query_qk[0])

    key_qk, key_qb = op.key_dense.qkernel, op.key_dense.qbias
    key_qk = to_np_arr(key_qk)
    key_qb = to_np_arr(key_qb) if key_qb is not None else np.zeros_like(key_qk[0])

    value_qk, value_qb = op.value_dense.qkernel, op.value_dense.qbias
    value_qk = to_np_arr(value_qk)
    value_qb = to_np_arr(value_qb) if value_qb is not None else np.zeros_like(value_qk[0])

    if op._fuse == 'qkv':
        assert query is key and key is value, 'Fused QKV projection only works when query, key and value are the same.'
        iq = op.query_dense.iq if op.enable_iq else None
        query = mirror_quantizer(iq, query) if iq is not None else query

        to_QKV_kernel = np.concatenate([query_qk, key_qk, value_qk], axis=-1)
        to_QKV_bias = np.concatenate([query_qb, key_qb, value_qb], axis=-1)

        QKV = einsum(op.query_dense.equation, query[None], to_QKV_kernel)[0] + to_QKV_bias
        Q, K, V = np.split(QKV, 3, axis=-1)  # type: ignore

        Q = mirror_quantizer(op._query_dense.oq, Q) if op._query_dense.oq else Q
        K = mirror_quantizer(op._key_dense.oq, K) if op._key_dense.enable_oq else K
        V = mirror_quantizer(op._value_dense.oq, V) if op._value_dense.enable_oq else V

    elif op._fuse == 'kv':
        assert key is value, 'Fused KV projection only works when key and value are the same.'
        iq = op._key_dense.iq if op.enable_iq else None
        key = mirror_quantizer(iq, key) if iq is not None else key

        to_KV_kernel = np.concatenate([key_qk, value_qk], axis=-1)
        to_KV_bias = np.concatenate([key_qb, value_qb], axis=-1)
        KV = einsum(op._key_dense.equation, key[None], to_KV_kernel)[0] + to_KV_bias
        K, V = np.split(KV, 2, axis=-1)  # type: ignore

        K = mirror_quantizer(op._key_dense.oq, K) if op._key_dense.enable_oq else K
        V = mirror_quantizer(op._value_dense.oq, V) if op._value_dense.enable_oq else V

        q_query = mirror_quantizer(op._query_dense.iq, query) if op._query_dense.iq is not None else query
        Q = einsum(op._query_dense.equation, q_query[None], query_qk)[0] + query_qb
        Q = mirror_quantizer(op._query_dense.oq, Q) if op._query_dense.enable_oq else Q

    else:
        q_query = mirror_quantizer(op._query_dense.iq, query) if op._query_dense.iq is not None else query
        Q = einsum(op._query_dense.equation, q_query[None], query_qk)[0] + query_qb
        Q = mirror_quantizer(op._query_dense.oq, Q) if op._query_dense.enable_oq else Q

        q_key = mirror_quantizer(op._key_dense.iq, key) if op._key_dense.iq is not None else key
        K = einsum(op._key_dense.equation, q_key[None], key_qk)[0] + key_qb
        K = mirror_quantizer(op._key_dense.oq, K) if op._key_dense.enable_oq else K

        q_value = mirror_quantizer(op._value_dense.iq, value) if op._value_dense.iq is not None else value
        V = einsum(op._value_dense.equation, q_value[None], value_qk)[0] + value_qb
        V = mirror_quantizer(op._value_dense.oq, V) if op._value_dense.enable_oq else V

    return Q, K, V


class ReplayMHA(ReplayOperationBase):
    handles = (QMultiHeadAttention,)
    __input_quantizer_handled__ = True
    __output_quantizer_handled__ = True

    def _compute_attention_mask(
        self,
        query,
        value,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        use_causal_mask=False,
    ):
        masks = []
        if query_mask is not None:
            masks.append(np.expand_dims(query_mask, -1))  # [Q, 1]
        if value_mask is not None:
            masks.append(np.expand_dims(value_mask, -2))  # [1, V]
        if key_mask is not None:
            masks.append(np.expand_dims(key_mask, -2))  # [1, V]
        if use_causal_mask:
            q = query.shape[0]
            v = q if value is None else value.shape[0]
            masks.append(np.tril(np.ones((q, v), dtype='uint8')))  # [Q, V]
        if attention_mask is not None:
            masks.append(attention_mask)
        if not masks:
            return None

        return np.prod(np.stack(masks, axis=0), axis=0)

    def _masked_softmax(self, op, attention_scores, attention_mask=None):
        # Normalize the attention scores to probabilities.
        # attention_scores = [B, N, T, S]
        if attention_mask is not None:
            # The expand dim happens starting from the `num_heads` dimension,
            # (<batch_dims>, num_heads, <query_attention_dims,
            # key_attention_dims>)
            mask_expansion_axis = -len(op._attention_axes) * 2 - 1
            for _ in range(len(attention_scores.shape) - len(attention_mask.shape)):
                attention_mask = np.expand_dims(attention_mask, axis=mask_expansion_axis)
            attention_mask = attention_mask[0]

        return ReplayQSoftmax(op._softmax)(attention_scores[0], mask=attention_mask)[0][None]

    def _compute_attention(self, op: QMultiHeadAttention, query, key, value, attention_mask=None, training=None):
        # Take the dot product between "query" and "key" to get the raw
        # attention scores.
        attention_scores = einsum(op._dot_product_equation, key, query)

        attention_scores = self._masked_softmax(op, attention_scores, attention_mask)

        # `context_layer` = [B, T, N, H]
        attention_output = einsum(op._combine_equation, attention_scores, value)
        return attention_output, attention_scores

    def call(
        self,
        query: FixedVariableArray,
        value=None,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        use_causal_mask=False,
    ):
        op: QMultiHeadAttention = self.op

        value = query if value is None else value
        key = value if key is None else key

        _attention_mask = self._compute_attention_mask(
            query,
            value,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            use_causal_mask=use_causal_mask,
        )

        query, key, value = _fused_qkv_proj(query, key, value, op)
        query, key, value = query[None], key[None], value[None]

        attention_output, attention_scores = self._compute_attention(op, query, key, value, _attention_mask)
        attention_output = ReplayQDense(op._output_dense)(attention_output[0])[0]

        if op.enable_oq:
            attention_output = mirror_quantizer(op.oq, attention_output)

        if return_attention_scores:
            return attention_output, attention_scores[0]
        return attention_output


class ReplayQLinformerAttention(ReplayMHA):
    handles = (QLinformerAttention,)

    def call(
        self,
        query,
        value=None,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        use_causal_mask=False,
    ):
        assert use_causal_mask is False, 'Causal mask is not supported in QLinformerAttention.'
        value = value if value is not None else query
        key = key if key is not None else value
        op: QLinformerAttention = self.op
        key = ReplayQDense(op._lin_k_proj)(key)[0]
        value = ReplayQDense(op._lin_v_proj)(value)[0]
        return super().call(
            query,
            value,
            key,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            return_attention_scores=return_attention_scores,
        )


class ReplaySALTAttention(ReplayMHA):
    handles = (QSALTAttention,)

    def _masked_softmax(self, op, attention_scores, attention_mask=None):
        self.op: QSALTAttention
        if self.op.conv_size > 0:
            attention_scores = ReplayQConv(self.op.conv)(attention_scores[0])[0][None]
        return super()._masked_softmax(op, attention_scores, attention_mask)

    def call(
        self,
        query,
        value=None,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        use_causal_mask=False,
    ):
        assert use_causal_mask is False, 'Causal mask is not supported in QLinformerAttention.'
        value = value if value is not None else query
        key = key if key is not None else value
        op = self.op

        if self.op.cluster_k_proj:
            key = np.pad(key, [[0, self.op.n_k_pad], [0, 0]], mode='constant', constant_values=0)  # type: ignore
            key = np.reshape(key, self.op._k_reshape_to[1:])  # type: ignore

        if self.op.n_v_pad > 0:
            value = np.pad(value, [[0, self.op.n_v_pad], [0, 0]], mode='constant', constant_values=0)  # type: ignore
            value = np.reshape(value, self.op._v_reshape_to[1:])  # type: ignore
        key = ReplayQDense(op._lin_k_proj)(key)[0]
        value = ReplayQDense(op._lin_v_proj)(value)[0]
        return super().call(
            query,
            value,
            key,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            return_attention_scores=return_attention_scores,
        )


__all__ = ['ReplayMHA', 'ReplayQLinformerAttention', 'ReplaySALTAttention']



from typing import Sequence

from hgq import layers
from hgq.config import QuantizerConfig

class QDense(layers.QDense):
    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kernel_quantizer: None | QuantizerConfig | str = None,
        input_quantizer: None | QuantizerConfig | str = None,
        bias_quantizer: None | QuantizerConfig | str = None,
        output_quantizer: None | QuantizerConfig | str = None,
        **kwargs,
    ): ...

class QEinsumDense(layers.QEinsumDense):
    def __init__(
        self,
        equation,
        output_shape,
        activation=None,
        bias_axes=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kernel_quantizer: None | QuantizerConfig | str = None,
        input_quantizer: None | QuantizerConfig | str = None,
        bias_quantizer: None | QuantizerConfig | str = None,
        output_quantizer: None | QuantizerConfig | str = None,
        **kwargs,
    ): ...

class QConv1D(layers.QConv1D):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding='valid',
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kernel_quantizer: None | QuantizerConfig | str = None,
        input_quantizer: None | QuantizerConfig | str = None,
        bias_quantizer: None | QuantizerConfig | str = None,
        output_quantizer: None | QuantizerConfig | str = None,
        **kwargs,
    ): ...

class QConv2D(layers.QConv2D):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1),
        padding='valid',
        data_format=None,
        dilation_rate=(1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kernel_quantizer: None | QuantizerConfig | str = None,
        input_quantizer: None | QuantizerConfig | str = None,
        bias_quantizer: None | QuantizerConfig | str = None,
        output_quantizer: None | QuantizerConfig | str = None,
        **kwargs,
    ): ...

class QConv3D(layers.QConv3D):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1, 1),
        padding='valid',
        data_format=None,
        dilation_rate=(1, 1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kernel_quantizer: None | QuantizerConfig | str = None,
        input_quantizer: None | QuantizerConfig | str = None,
        bias_quantizer: None | QuantizerConfig | str = None,
        output_quantizer: None | QuantizerConfig | str = None,
        **kwargs,
    ): ...

class QSoftmax(layers.QSoftmax):
    def __init__(
        self,
        axis: int | Sequence[int] = -1,
        iq_conf: None | QuantizerConfig = None,
        stable=False,
        exp_input_quantizer: None | QuantizerConfig | str = None,
        exp_output_quantizer: None | QuantizerConfig | str = None,
        inv_input_quantizer: None | QuantizerConfig | str = None,
        inv_output_quantizer: None | QuantizerConfig | str = None,
        allow_heterogeneous_table: bool = False,
        input_scaler: float = 1.0,
        **kwargs,
    ): ...

class QActivation(layers.QUnaryFunctionLUT):
    def __init__(
        self,
        activation,
        input_quantizer: None | QuantizerConfig | str = None,
        output_quantizer: None | QuantizerConfig | str = None,
        allow_heterogeneous_table: bool = False,
        **kwargs,
    ): ...

class QMeanPow2(layers.QMeanPow2):
    pass

class QSum(layers.QSum):
    pass

class QAdd(layers.QAdd):
    pass

class QAveragePow2(layers.QAveragePow2):
    pass

class QDot(layers.QDot):
    pass

class QEinsum(layers.QEinsum):
    pass

class QMaximum(layers.QMaximum):
    pass

class QMinimum(layers.QMinimum):
    pass

class QMultiply(layers.QMultiply):
    pass

class QSubtract(layers.QSubtract):
    pass



from collections.abc import Sequence

import numpy as np
from keras import ops
from keras.saving import register_keras_serializable
from keras.src import backend
from keras.src.layers import Layer

numbers = int | float | np.integer | np.floating


def _large_number(dtype):
    """Return a Large negative number based on dtype."""
    if backend.standardize_dtype(dtype) == 'float16':
        return 3e4
    return 1e9


@ops.custom_gradient
def round_conv(x):
    qx = ops.round(x)

    def grad(*args, upstream=None):
        if upstream is None:
            (upstream,) = args
        return upstream

    return qx, grad


class BitwidthMapperBase:
    """Abstract base class for mapping bitwidth tensor to input tensors for HG quantizers."""

    def bw_to_x(self, bw, x_shape):
        raise NotImplementedError

    def x_to_bw_absmax(self, x):
        raise NotImplementedError

    def inference_weight_shape(self, input_shape) -> tuple[int, ...]:
        raise NotImplementedError

    def get_config(self):
        raise NotImplementedError

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def check_axis(axis: Sequence[int], ndim: int):
    """Given a list of axis, check that they are valid for a tensor of ndim dimensions. If valid, return the axis as a list of positive integers.

    Parameters
    ----------
    axis : Sequence[int]
        List of axis to check.
    ndim : int
        Number of dimensions of the tensor.

    Returns
    -------
    axis : list[int]
        List of positive integers representing the axis.
    """
    axis = [a if a >= 0 else a + ndim for a in axis]
    assert all(0 <= a < ndim for a in axis), f'Invalid axis {axis} for shape {ndim}.'
    return axis


class TrainableQuantizerBase(Layer):
    """Abstract base class for all quantizers."""

    __dummy__ = False

    def __init__(self, **kwargs):
        homogeneous_axis = kwargs.pop('homogeneous_axis', ())
        heterogeneous_axis = kwargs.pop('heterogeneous_axis', None)
        bw_mapper: BitwidthMapperBase = kwargs.pop('bw_mapper', None) or DefaultBitwidthMapper(
            heterogeneous_axis, homogeneous_axis
        )
        self.bw_mapper = bw_mapper
        self._seed = kwargs.pop('seed', int(np.random.randint(0, 2**31)))
        super().__init__(**kwargs)
        self.supports_masking = True

    def call(self, inputs, training=None): ...

    def __repr__(self) -> str: ...

    def quantize(self, mode):  # type: ignore
        raise ValueError('Quantize method is built-in for keras v3. This method is disabled in this package.')

    def compute_output_shape(self, input_shape):
        return input_shape

    @property
    def bits(self):
        raise NotImplementedError

    @property
    def fbits(self):
        raise NotImplementedError

    @property
    def min(self):
        raise NotImplementedError

    @property
    def max(self):
        raise NotImplementedError

    @property
    def epsilon(self):
        raise NotImplementedError


class DummyQuantizer(TrainableQuantizerBase):
    __dummy__ = True

    def call(self, inputs, training=None):
        return inputs

    @property
    def bits(self):
        return ops.convert_to_tensor(0, 'float32')

    @property
    def fbits(self):
        return ops.convert_to_tensor(0, 'float32')

    @property
    def min(self):
        return -_large_number(self.dtype)

    @property
    def max(self):
        return _large_number(self.dtype)

    @property
    def epsilon(self):
        return backend.epsilon()

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name}, built={self.built})'


@register_keras_serializable(package='hgq')
class DefaultBitwidthMapper(BitwidthMapperBase):
    """Default bitwidth mapper for HG quantizers."""

    def __init__(self, heterogeneous_axis: Sequence[int] | None = None, homogeneous_axis: Sequence[int] | None = None, **kwargs):
        _shape_inferenced = kwargs.pop('_shape_inferenced', False)
        if not _shape_inferenced:
            assert (heterogeneous_axis is None) ^ (homogeneous_axis is None), (
                'Only one of quantize_dims and skip_dims can be specified.'
            )
        self.heterogeneous_axis = heterogeneous_axis
        self.homogeneous_axis = homogeneous_axis
        self._shape_inferenced = _shape_inferenced
        super().__init__(**kwargs)

    def inference_weight_shape(self, input_shape):
        N = len(input_shape)
        axis = np.arange(N)
        if self.heterogeneous_axis is not None:
            self.heterogeneous_axis = check_axis(self.heterogeneous_axis, N)  # type: ignore
            self.homogeneous_axis = tuple(np.setdiff1d(axis, self.heterogeneous_axis))
        elif self.homogeneous_axis is not None:
            self.homogeneous_axis = check_axis(self.homogeneous_axis, N)  # type: ignore
            self.heterogeneous_axis = tuple(np.setdiff1d(axis, self.homogeneous_axis))

        weight_shape = [1] * N
        for i in self.heterogeneous_axis:  # type: ignore
            assert input_shape[i] is not None, (
                f'Unable to heterogeneously quantize axis {i} with unknown shape. Input shape: {input_shape}.'
            )
            weight_shape[i] = input_shape[i]
        self._shape_inferenced = True
        return tuple(weight_shape)

    def bw_to_x(self, bw, x_shape):
        return ops.broadcast_to(bw, x_shape)

    def x_to_bw_absmax(self, x):
        return ops.max(ops.abs(x), axis=self.homogeneous_axis, keepdims=True)

    def x_to_bw_sign(self, x):
        return ops.any(x < 0, axis=self.homogeneous_axis, keepdims=True)

    def get_config(self):
        return dict(
            heterogeneous_axis=self.heterogeneous_axis,
            homogeneous_axis=self.homogeneous_axis,
            _shape_inferenced=self._shape_inferenced,
        )



import math
from collections.abc import Sized
from typing import Literal

from keras import ops
from keras.initializers import Constant
from keras.layers import Dropout, MultiHeadAttention
from keras.src.layers.attention.multi_head_attention import _build_attention_equation, _build_proj_equation

from ...quantizer.config import QuantizerConfig
from ...utils.misc import gather_vars_to_kwargs
from ..core.base import QLayerBase
from ..core.einsum_dense import QEinsumDense
from ..softmax import QSoftmax


def _get_output_shape(output_rank, known_last_dims, input_shape):
    n = output_rank - len(known_last_dims)
    return list(input_shape[1 : n + 1]) + list(known_last_dims)


class QMultiHeadAttention(MultiHeadAttention, QLayerBase):
    __output_quantizer_handled__ = True

    def __init__(
        self,
        num_heads,
        key_dim,
        value_dim=None,
        dropout=0.0,
        use_bias=True,
        output_shape=None,
        attention_axes=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        seed=None,
        fuse: Literal['none', 'qkv', 'kv'] = 'none',
        qkvo_iq_conf: QuantizerConfig | None = None,
        qkvo_kq_conf: QuantizerConfig | None = None,
        qkvo_bq_conf: QuantizerConfig | None = None,
        qkvo_oq_conf: QuantizerConfig | None = None,
        softmax_iq_conf: QuantizerConfig | None = None,
        softmax_exp_iq_conf: QuantizerConfig | None = None,
        softmax_exp_oq_conf: QuantizerConfig | None = None,
        softmax_inv_iq_conf: QuantizerConfig | None = None,
        softmax_inv_oq_conf: QuantizerConfig | None = None,
        softmax_oq_conf: QuantizerConfig | None = None,
        stable_softmax=True,
        softmax_allow_heterogeneous_table: bool = False,
        parallelization_factor=-1,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self|.+q_conf')

        self._qkvo_iq_conf = qkvo_iq_conf or QuantizerConfig(place='datalane')
        self._qkvo_kq_conf = qkvo_kq_conf or QuantizerConfig(place='weight')
        self._qkvo_bq_conf = qkvo_bq_conf or QuantizerConfig(place='bias')
        self._qkvo_oq_conf = qkvo_oq_conf or QuantizerConfig(place='datalane')
        self._softmax_iq_conf = softmax_iq_conf or QuantizerConfig(place='datalane')
        self._softmax_exp_iq_conf = softmax_exp_iq_conf or QuantizerConfig(place='datalane')
        self._softmax_exp_oq_conf = softmax_exp_oq_conf or QuantizerConfig(place='table')
        self._softmax_inv_iq_conf = softmax_inv_iq_conf or QuantizerConfig(place='datalane')
        self._softmax_inv_oq_conf = softmax_inv_oq_conf or QuantizerConfig(place='table')
        self._softmax_oq_conf = softmax_oq_conf or QuantizerConfig(place='datalane')
        self._softmax_allow_heterogeneous_table = kwargs.pop('softmax_allow_heterogeneous_table')
        self.parallelization_factor = kwargs.pop('parallelization_factor')
        self._stable_softmax = kwargs.pop('stable_softmax')
        self._fuse = kwargs.pop('fuse', 'none').lower()
        self._inverse_sqrt_key_dim = 1.0 / math.sqrt(float(key_dim))

        super().__init__(**kwargs)

    def _get_common_kwargs_for_sublayer(self):
        common_kwargs: dict = super()._get_common_kwargs_for_sublayer()
        # Inject quantizer and ebops configs to sub QEinsumDense layers.
        common_kwargs.update(
            {
                'iq_conf': self._qkvo_iq_conf,
                'kq_conf': self._qkvo_kq_conf,
                'bq_conf': self._qkvo_bq_conf,
                'oq_conf': self._qkvo_oq_conf,
                'enable_ebops': self.enable_ebops,
                'beta0': self._beta0.clone(),
                'parallelization_factor': self.parallelization_factor,
            }
        )
        return common_kwargs

    def compute_output_spec(self, query, value=None, key=None, *args, **kwargs):
        value = query if value is None else value
        return super().compute_output_spec(*args, query=query, value=value, key=key, **kwargs)

    def build(
        self,
        query_shape,
        value_shape=None,
        key_shape=None,
    ):
        """Builds layers and variables.

        Parameters
        ----------
        query_shape : tuple
            Shape of the `query` tensor.
        value_shape : tuple, optional
            Shape of the `value` tensor.
        key_shape : tuple, optional
            Shape of the `key` tensor.
        """

        # Copied and modified from keras MultiHeadAttention, substituted
        # EinsumDense with QEinsumDense and added sequence length (shape) to its
        # output shape when initializing, if known.
        value_shape = query_shape if value_shape is None else value_shape
        key_shape = value_shape if key_shape is None else key_shape

        if self._fuse == 'qkv':
            assert query_shape == value_shape == key_shape, (
                'For fuse=qkv, query, key and value must have the same shape',
                f'but got query_shape={query_shape}, value_shape={value_shape} and key_shape={key_shape}.',
            )
        if self._fuse == 'kv':
            assert value_shape == key_shape, (
                'For fuse=kv, key and value must have the same shape',
                f'but got value_shape={value_shape} and key_shape={key_shape}.',
            )

        if value_shape[1:-1] != key_shape[1:-1]:
            raise ValueError(
                'All dimensions of `value` and `key`, except the last one, '
                f'must be equal. Received: value_shape={value_shape} and '
                f'key_shape={key_shape}',
            )

        query_rank = len(query_shape)
        key_rank = len(key_shape)
        value_rank = len(value_shape)

        einsum_equation, bias_axes, output_rank = _build_proj_equation(
            query_rank - 1,
            bound_dims=1,
            output_dims=2,
        )
        self._query_dense = QEinsumDense(
            einsum_equation,
            output_shape=_get_output_shape(
                output_rank - 1,
                [self._num_heads, self._key_dim],
                query_shape,
            ),
            bias_axes=bias_axes if self._use_bias else None,
            name='query',
            enable_iq=self.enable_iq and not self._fuse == 'qkv',
            enable_oq=True,
            **self._get_common_kwargs_for_sublayer(),
        )

        einsum_equation, bias_axes, output_rank = _build_proj_equation(
            key_rank - 1,
            bound_dims=1,
            output_dims=2,
        )
        self._key_dense = QEinsumDense(
            einsum_equation,
            output_shape=_get_output_shape(
                output_rank - 1,
                [self._num_heads, self._key_dim],
                key_shape,
            ),
            bias_axes=None,  # Useless as it will be directly fed to softmax on seq axis
            name='key',
            enable_iq=self.enable_iq and self._fuse not in ('qkv', 'kv'),
            enable_oq=True,
            **self._get_common_kwargs_for_sublayer(),
        )

        einsum_equation, bias_axes, output_rank = _build_proj_equation(
            value_rank - 1,
            bound_dims=1,
            output_dims=2,
        )
        self._value_dense = QEinsumDense(
            einsum_equation,
            output_shape=_get_output_shape(
                output_rank - 1,
                [self._num_heads, self._value_dim],
                value_shape,
            ),
            bias_axes=bias_axes if self._use_bias else None,
            name='value',
            enable_iq=self.enable_iq,
            enable_oq=True,
            **self._get_common_kwargs_for_sublayer(),
        )
        self._value_dense.build(value_shape)

        if self._fuse == 'qkv' and self.enable_iq:
            self._query_dense._iq = self._value_dense._iq
            self._query_dense._enable_iq = True
        if self._fuse in ('qkv', 'kv') and self.enable_iq:
            self._key_dense._iq = self._value_dense._iq
            self._key_dense._enable_iq = True

        self._query_dense.build(query_shape)
        self._key_dense.build(key_shape)

        # Builds the attention computations for multi-head dot product
        # attention.  These computations could be wrapped into the keras
        # attention layer once it supports multi-head einsum computations.
        self._build_attention(output_rank, (query_shape, value_shape, key_shape))
        self._output_dense = self._make_output_dense(
            query_shape,
            self._get_common_kwargs_for_sublayer(),
            'attention_output',
        )
        output_dense_input_shape = list(
            self._query_dense.compute_output_shape(query_shape),
        )
        output_dense_input_shape[-1] = self._value_dim
        self._output_dense.build(tuple(output_dense_input_shape))

        if self.enable_ebops:
            self._beta = self.add_weight(
                name='beta',
                shape=(),
                initializer=self._beta0,
                trainable=False,
            )
            self._ebops = self.add_weight(
                name='ebops',
                shape=(),
                initializer=Constant(0.0),
                trainable=False,
                dtype='uint32',
            )
        else:
            self._beta = None
            self._ebops = None

        self._dot_product_ebops_equation = self._dot_product_equation.split('->', 1)[0] + '->'
        self._combine_ebops_equation = self._combine_equation.split('->', 1)[0] + '->'
        self.built = True

    def _make_output_dense(self, query_shape, common_kwargs, name=None):
        """Builds the output projection matrix.

        Parameters
        ----------
        query_shape : tuple
            Shape of the query tensor.
        common_kwargs : dict
            Common keyword arguments for the einsum layer.
        name : str, optional
            Name for the projection layer.

        Returns
        -------
        QEinsumDense

        Notes
        -----
        This method is copied and modified from Keras MultiHeadAttention. It substitutes
        EinsumDense with QEinsumDense and adds sequence length (shape) to its output shape
        when initializing, if known.
        """

        query_rank = len(query_shape)
        if self._output_shape:
            if not isinstance(self._output_shape, Sized):
                output_shape = [self._output_shape]
            else:
                output_shape = self._output_shape
        else:
            output_shape = [query_shape[-1]]
        einsum_equation, bias_axes, output_rank = _build_proj_equation(
            query_rank - 1,
            bound_dims=2,
            output_dims=len(output_shape),
        )
        return QEinsumDense(
            einsum_equation,
            output_shape=_get_output_shape(output_rank - 1, output_shape, query_shape),
            bias_axes=bias_axes if self._use_bias else None,
            name=name,
            enable_iq=True,
            enable_oq=self.enable_oq,
            **common_kwargs,
        )

    def _build_attention(self, rank, shapes=None):
        """Builds multi-head dot-product attention computations.

        This function builds attributes necessary for `_compute_attention` to
        customize attention computation to replace the default dot-product
        attention.

        Parameters
        ----------
        rank : int
            The rank of query, key, value tensors.
        """

        # Copied and modified from keras MultiHeadAttention, substituted Softmax with QSoftmax.
        if self._attention_axes is None:
            self._attention_axes = tuple(range(1, rank - 2))
        elif not isinstance(self._attention_axes, tuple):
            self._attention_axes = tuple(self._attention_axes)
        (
            self._dot_product_equation,
            self._combine_equation,
            attn_scores_rank,
        ) = _build_attention_equation(rank, attn_axes=self._attention_axes)
        norm_axes = tuple(
            range(
                attn_scores_rank - len(self._attention_axes),
                attn_scores_rank,
            ),
        )
        self._softmax = QSoftmax(
            enable_oq=True,
            axis=norm_axes,
            dtype=self.dtype_policy,
            stable=self._stable_softmax,
            iq_conf=self._softmax_iq_conf,
            exp_iq_conf=self._softmax_exp_iq_conf,
            exp_oq_conf=self._softmax_exp_oq_conf,
            inv_iq_conf=self._softmax_inv_iq_conf,
            inv_oq_conf=self._softmax_inv_oq_conf,
            oq_conf=self._softmax_oq_conf,
            allow_heterogeneous_table=self._softmax_allow_heterogeneous_table,
            input_scaler=self._inverse_sqrt_key_dim,
            enable_ebops=self.enable_ebops,
        )
        self._dropout_layer = Dropout(
            rate=self._dropout,
            dtype=self.dtype_policy,
            seed=self.seed,
        )
        self._inverse_sqrt_key_dim = 1.0
        # Build softmax and dropout layers if possible.
        if shapes is not None:
            q_shape, v_shape, _ = shapes
            attn_score_shape = (None, self._num_heads, *q_shape[1:-1], *v_shape[1:-1])
            self._softmax.build(attn_score_shape)
            self._dropout_layer.build(attn_score_shape)

    def compute_output_shape(self, query_shape, value_shape, key_shape=None):
        return super().compute_output_shape(query_shape, query_shape, None)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'qkvo_iq_conf': self._qkvo_iq_conf,
                'qkvo_kq_conf': self._qkvo_kq_conf,
                'qkvo_bq_conf': self._qkvo_bq_conf,
                'qkvo_oq_conf': self._qkvo_oq_conf,
                'softmax_iq_conf': self._softmax_iq_conf,
                'softmax_exp_iq_conf': self._softmax_exp_iq_conf,
                'softmax_exp_oq_conf': self._softmax_exp_oq_conf,
                'softmax_inv_iq_conf': self._softmax_inv_iq_conf,
                'softmax_inv_oq_conf': self._softmax_inv_oq_conf,
                'softmax_oq_conf': self._softmax_oq_conf,
                'softmax_allow_heterogeneous_table': self._softmax_allow_heterogeneous_table,
                'parallelization_factor': self.parallelization_factor,
                'stable_softmax': self._stable_softmax,
                'fuse': self._fuse,
            }
        )
        return config

    def _post_build(self):
        if self._enable_oq:
            assert hasattr(self, '_oq'), f'Output Quantizer is not defined for {self.name}, but enable_oq is True.'
        for sublayer in self._flatten_layers():
            assert sublayer.built, f'Sublayer {sublayer.name} is not built for {self.name}'

    def _compute_ebops(self, query_shape, value_shape=None, key_shape=None):
        Q_shape = (1,) + self._query_dense.full_output_shape[1:]
        K_shape = (1,) + self._key_dense.full_output_shape[1:]
        V_shape = (1,) + self._value_dense.full_output_shape[1:]

        value_shape = query_shape if value_shape is None else value_shape
        attn_score_shape = (1, self._num_heads, *query_shape[1:-1], *value_shape[1:-1])

        # PF not supported for MHA for now.
        # if self.parallelization_factor > 0:
        #     assert len(query_shape) == 3, f'EBOPs computation with pf>0 is only supported for 3D tensors, but got {query_shape}.'
        #     b, *n, h, dk = Q_shape
        #     b, *n, h, dv = K_shape
        #     b, *n, h, dv = V_shape

        #     Q_shape = b, (1,) * len(n), h, dk
        #     K_shape = b, (1,) * len(n), h, dv
        #     V_shape = b, (1,) * len(n), h, dv
        #     attn_score_shape = b, self._num_heads, *(1,) * len(n) * 2

        bw_q = self._query_dense.oq.bits_(Q_shape)
        bw_k = self._key_dense.oq.bits_(K_shape)
        bw_v = self._value_dense.oq.bits_(V_shape)
        bw_attn = self._softmax.oq.bits_(attn_score_shape)

        ebops_qk = ops.einsum(self._dot_product_ebops_equation, bw_q, bw_k)
        ebops_av = ops.einsum(self._combine_ebops_equation, bw_attn, bw_v)
        ebops = ebops_qk + ebops_av  # type: ignore
        if self.parallelization_factor > 0:
            return ebops * self.parallelization_factor
        return ebops

    @property
    def ebops(self):
        if self._ebops is None:
            return ops.cast(0, 'uint32')
        ebops = sum(
            (  # type: ignore
                self._query_dense.ebops,
                self._key_dense.ebops,
                self._value_dense.ebops,
                self._softmax.ebops,
                self._output_dense.ebops,
                ops.convert_to_tensor(self._ebops),  # type: ignore
            )
        )
        return ebops  # type: ignore

    def call(
        self,
        query,
        value=None,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        training=None,
        use_causal_mask=False,
    ):
        # Adapted from _compute_attention in keras 3.5.0

        if value is None:
            value = query
        if key is None:
            key = value

        attention_mask = self._compute_attention_mask(
            query,
            value,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            use_causal_mask=use_causal_mask,
        )

        #   N = `num_attention_heads`
        #   H = `size_per_head`
        # `query` = [B, T, N ,H]
        query = self._query_dense(query)

        # `key` = [B, S, N, H]
        key = self._key_dense(key)

        # `value` = [B, S, N, H]
        value = self._value_dense(value)

        attention_output, attention_scores = self._compute_attention(query, key, value, attention_mask, training)
        attention_output = self._output_dense(attention_output)

        if self.enable_oq:
            attention_output = self.oq(attention_output, training=training)

        if return_attention_scores:
            return attention_output, attention_scores
        return attention_output

    def _compute_attention(self, query, key, value, attention_mask=None, training=None):  # type: ignore
        # Original _compute_attention in keras 3.5.0
        # Copied for disable to flash-attn that breaks quantization.
        """Applies Dot-product attention with query, key, value tensors.

        This function defines the computation inside `call` with projected
        multi-head Q, K, V inputs. Users can override this function for
        customized attention implementation.

        Parameters
        ----------
        query : tensor
            Projected query tensor of shape `(B, T, N, key_dim)`.
        key : tensor
            Projected key tensor of shape `(B, S, N, key_dim)`.
        value : tensor
            Projected value tensor of shape `(B, S, N, value_dim)`.
        attention_mask : tensor, optional
            A boolean mask of shape `(B, T, S)` that prevents attention to
            certain positions. It is generally not needed if the `query` and
            `value` (and/or `key`) are masked.
        training : bool, optional
            Python boolean indicating whether the layer should behave in
            training mode (adding dropout) or in inference mode (doing nothing).

        Returns
        -------
        attention_output : tensor
            Multi-headed outputs of attention computation.
        attention_scores : tensor
            Multi-headed attention weights.
        """
        # Note: Applying scalar multiply at the smaller end of einsum improves
        # XLA performance, but may introduce slight numeric differences in
        # the Transformer attention head.
        query = ops.multiply(query, ops.cast(self._inverse_sqrt_key_dim, query.dtype))

        # Take the dot product between "query" and "key" to get the raw
        # attention scores.
        attention_scores = ops.einsum(self._dot_product_equation, key, query)

        attention_scores = self._masked_softmax(attention_scores, attention_mask)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        if self.dropout:
            final_attn_scores = self._dropout_layer(attention_scores, training=training)
        else:
            final_attn_scores = attention_scores

        # `context_layer` = [B, T, N, H]
        attention_output = ops.einsum(self._combine_equation, final_attn_scores, value)
        return attention_output, attention_scores



from collections.abc import Callable
from math import prod, sqrt

import keras
from keras import ops
from keras.layers import Layer
from keras.saving import register_keras_serializable

from ...quantizer import Quantizer, QuantizerConfig
from ..core import QLayerBaseSingleInput


@register_keras_serializable(package='hgq2')
class QDenseT(QLayerBaseSingleInput):
    def __init__(
        self,
        n_out: int,
        n_hl: int = 1,
        d_hl: int = 8,
        activation: Callable | None | str = None,
        subnn_activation: str | Callable | Layer | None = 'tanh',
        toq_conf: QuantizerConfig | None = None,
        parallelization_factor: int = -1,
        use_bias: bool = True,
        batch_norm: bool = False,
        bn_center: bool = True,
        bn_scale: bool = True,
        bn_momentum: float = 0.99,
        bn_epsilon: float = 0.001,
        table_spec: int | tuple[int, int] = (6, 5),
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.n_out = n_out
        self.d_hl = d_hl
        self.use_bias = use_bias
        self.subnn_activation = keras.activations.get(subnn_activation)
        self.parallelization_factor = parallelization_factor
        self.enable_bn = batch_norm
        self.table_spec = (table_spec, table_spec) if isinstance(table_spec, int) else table_spec
        self.bn_args = {
            'center': bn_center,
            'scale': bn_scale,
            'momentum': bn_momentum,
            'epsilon': bn_epsilon,
        }
        assert n_hl >= 0

        self.n_hl = n_hl

        toq_conf = toq_conf or QuantizerConfig(place='table')
        self._toq = Quantizer(toq_conf)
        self.activation = keras.activations.get(activation)

    def _build_module(self, n_in: int):
        layers = []
        _shape = (n_in, self.n_out, self.d_hl)
        bias_axes = 'ioD' if self.use_bias else None
        for _ in range(self.n_hl):
            layers.append(
                keras.layers.EinsumDense(
                    '...iod,iodD->...ioD',
                    _shape,
                    self.subnn_activation,
                    bias_axes=bias_axes,
                )
            )
            # layers.extend(
            #     [
            #         keras.layers.EinsumDense(
            #             '...iod,iodD->...ioD',
            #             _shape,
            #             bias_axes=bias_axes,
            #         ),
            #         # keras.layers.Reshape((n_in * self.n_out * self.d_hl,)),
            #         keras.layers.BatchNormalization(axis=-1),
            #         # keras.layers.Reshape((n_in, self.n_out, self.d_hl)),
            #         keras.layers.Activation(self.subnn_activation),
            #     ]
            # )
        bias_axes = 'io' if self.use_bias else None
        l_out = keras.layers.EinsumDense('...iod,iod->...io', (n_in, self.n_out), 'linear', bias_axes=bias_axes)
        layers.append(l_out)
        module = keras.models.Sequential(layers)
        return module

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self.n_in = input_shape[-1]

        self.n_parallel = prod(input_shape[1:-1])
        if self.parallelization_factor < 0:
            self.parallelization_factor = self.n_parallel

        if self.enable_iq and not self.iq.built:
            self.iq.build(input_shape + (self.n_out,))
        self.toq.build(input_shape + (self.n_out,))
        self.module = self._build_module(self.n_in)
        self.module.build(input_shape + (self.n_out, 1))

        if self.enable_bn:
            self.bn_module = keras.layers.BatchNormalization(
                axis=-1,
                **self.bn_args,
            )
            self.bn_module.build(input_shape + (self.n_out,))

        super().build(input_shape)

    def call(self, x, training=None):
        n_in = ops.shape(x)[-1]
        x = ops.broadcast_to(x[..., None], (*ops.shape(x)[:-1], n_in, self.n_out))  # (B, N_in, N_out)
        if self.enable_iq:
            x = self.iq(x)
        x = x[..., None]  # (B, ..., N_in, N_out, 1)
        table_out = self.module(x, training=training)  # (B, ..., N_in, N_out, 1) -> (B, ..., N_in, N_out)

        if self.enable_bn:
            table_out = self.bn_module(table_out, training=training) / sqrt(self.n_in)

        return self.activation(ops.sum(self.toq(table_out), axis=-2))

    def _compute_ebops(self, shape: tuple[int, ...]):
        q_shape = shape + (self.n_out,)
        bits_in = self.iq.fbits_(q_shape)
        bits_out = self.toq.fbits_(q_shape)

        B, b = self.table_spec

        small_lut_count = ops.where(bits_in >= b, 2 ** (bits_in - b), bits_in / b)  # type: ignore

        large_lut_count = ops.dot(ops.ravel(small_lut_count), ops.ravel(bits_out)) * 2 ** (b - B)  # type: ignore

        return (large_lut_count + ops.sum(bits_out)) * self.parallelization_factor / self.n_parallel

    @property
    def toq(self):
        return self._toq

    def get_config(self):
        config = {
            'n_out': self.n_out,
            'n_hl': self.n_hl,
            'd_hl': self.d_hl,
            'subnn_activation': self.subnn_activation,
            'activation': self.activation,
            'toq_conf': self.toq.config,
            'use_bias': self.use_bias,
            'parallelization_factor': self.parallelization_factor,
            'batch_norm': self.enable_bn,
            'bn_center': self.bn_args['center'],
            'bn_scale': self.bn_args['scale'],
            'bn_momentum': self.bn_args['momentum'],
            'bn_epsilon': self.bn_args['epsilon'],
            **super().get_config(),
        }
        return config



from keras.callbacks import Callback
from tqdm import tqdm


class PBar(Callback):
    def __init__(self, metric='loss: {loss:.2f}/{val_loss:.2f}', disable_ebops=False):
        self.pbar = None
        self.template = metric
        self.disable_ebops = disable_ebops

    def on_epoch_begin(self, epoch, logs=None):
        assert isinstance(self.params, dict)
        if self.pbar is None:
            self.pbar = tqdm(total=self.params['epochs'], unit='epoch')

    def on_epoch_end(self, epoch, logs=None):
        assert isinstance(self.pbar, tqdm)
        assert isinstance(logs, dict)
        self.pbar.update(1)
        string = self.template.format(**logs)
        if not self.disable_ebops and 'ebops' in logs:
            string += f' - EBOPs: {logs["ebops"]:,.0f}'
        self.pbar.set_description(string)

    def on_train_end(self, logs=None):
        if self.pbar is not None:
            self.pbar.close()
            self.pbar = None



from collections.abc import Callable
from math import prod, sqrt

import keras
import numpy as np
from da4ml.trace import FixedVariableArray
from da4ml.trace.fixed_variable import FixedVariable
from da4ml.trace.ops import _quantize
from keras import ops

from hgq.layers.table import QConvT1D, QConvT2D, QConvTBase, QDenseT
from hgq.quantizer.internal import FixedPointQuantizerBase

from ._base import ReplayOperationBase, mirror_quantizer, to_np_arr
from .conv import symbolic_extract_patches


def keras_act_to_numpy(act: Callable) -> Callable:
    match act:
        case keras.activations.relu:
            return lambda x: np.maximum(0, x)
        case keras.activations.tanh:
            return np.tanh
        case keras.activations.softmax:
            raise ValueError('Non-local activation must not be used')
        case keras.activations.linear:
            return lambda x: x
        case keras.activations.sigmoid:
            return lambda x: 1 / (1 + np.exp(-x))
        case keras.activations.swish:
            return lambda x: x / (1 + np.exp(-x))
        case keras.activations.gelu:
            return lambda x: 0.5 * x * (1 + np.tanh(np.sqrt(2 / np.pi) * (x + 0.044715 * np.power(x, 3))))
        case keras.activations.elu:
            return lambda x: np.where(x > 0, x, np.exp(x) - 1)
        case keras.activations.selu:
            alpha = 1.6732632423543772
            scale = 1.0507009873554805
            return lambda x: scale * np.where(x > 0, x, alpha * (np.exp(x) - 1))
        case keras.activations.softplus:
            return lambda x: np.log1p(np.exp(x))
        case keras.activations.softsign:
            return lambda x: x / (1 + np.abs(x))
        case keras.activations.exponential:
            return lambda x: np.exp(x)
        case keras.activations.hard_silu:
            return lambda x: x * np.minimum(1, np.maximum(0, (x + 1) / 2))
        case _:
            return lambda x: ops.convert_to_numpy(act(ops.convert_to_tensor(x)))


def gather_weights_and_activation(model: keras.Sequential):
    ws: list[np.ndarray] = []
    bs: list[np.ndarray | None] = []
    acts: list[Callable[[np.ndarray], np.ndarray]] = []
    for layer in model.layers:
        layer: keras.layers.EinsumDense
        w, *b = layer.get_weights()
        act = keras_act_to_numpy(layer.activation)
        if len(b) != 0:
            assert len(b) == 1
            b = b[0]
        else:
            b = None
        if w.ndim == 3:
            w = w[..., None]
            if b is not None:
                b = b[..., None]
        ws.append(w)
        bs.append(b)
        acts.append(act)
    return ws, bs, acts


class ReplayDenseTable(ReplayOperationBase):
    handles = (QDenseT,)

    __input_quantizer_handled__ = True

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        op: QDenseT = self.op  # type: ignore

        out = np.broadcast_to(inputs[..., None], inputs.shape + (op.n_out,))  # type: ignore
        out = mirror_quantizer(op.iq, out)

        l, h, s = out.lhs

        table_sizes: np.ndarray = np.round((h - l) / s).astype(np.uint32) + 1

        model = op.module

        ws, bs, acts = gather_weights_and_activation(model)

        out_shape: tuple[int, ...] = inputs.shape + (op.n_out,)
        tables: list[np.ndarray] = [None] * prod(out_shape)  # type: ignore
        n, loc = np.unique(table_sizes, return_inverse=True)

        for i in range(n.size):
            mask: np.ndarray = loc == i
            _l, _h = l[mask], h[mask]
            inp = np.linspace(_l, _h, n[i])

            _out = inp[..., None]

            idxs = np.where(mask.ravel())[0]
            mask = mask.reshape(-1, *mask.shape[-2:])

            for w, b, act in zip(ws, bs, acts):
                w = np.concatenate([w[_mask] for _mask in mask], axis=0)
                if b is not None:
                    b = np.concatenate([b[_mask] for _mask in mask], axis=0)
                else:
                    b = 0
                _out = act(np.einsum('...ni,nij->...nj', _out, w, optimize='optimal') + b)
            _out = _out[..., 0]

            for j, idx in enumerate(idxs):
                tables[idx] = _out[..., j]

        if op.enable_bn:
            bn = op.bn_module
            beta: np.ndarray = ops.convert_to_numpy(bn.beta) if bn.center else 1  # type: ignore
            gamma: np.ndarray = ops.convert_to_numpy(bn.gamma) if bn.scale else 1  # type: ignore
            m_mean: np.ndarray = ops.convert_to_numpy(bn.moving_mean)  # type: ignore
            m_var: np.ndarray = ops.convert_to_numpy(bn.moving_variance)  # type: ignore
            epsilon = bn.epsilon
            scaler = gamma / np.sqrt(m_var + epsilon)
            offset = beta - m_mean * scaler

            for i in range(len(tables)):
                tables[i][:] = (tables[i] * scaler[i % op.n_out] + offset[i % op.n_out]) / sqrt(op.n_in)

        assert all(v is not None for v in tables), tables

        toq = op.toq
        toq_internal: FixedPointQuantizerBase = toq.quantizer
        kk, ki, kf = toq_internal.kif

        _shape = (1,) + out.shape
        kk = toq_internal.bw_mapper.bw_to_x(kk, _shape)
        ki = toq_internal.bw_mapper.bw_to_x(ki, _shape)
        kf = toq_internal.bw_mapper.bw_to_x(kf, _shape)

        k, i, f = map(lambda x: to_np_arr(x).astype(np.int32).ravel(), (kk, ki, kf))

        round_mode, overflow_mode = toq_internal.round_mode, toq_internal.overflow_mode
        round_mode = round_mode[2:] if round_mode.startswith('S_') else round_mode
        for arr, _k, _i, _f in zip(tables, k, i, f):
            arr[:] = _quantize(arr, _k, _i, _f, overflow_mode, round_mode)

        ret_vars: list[FixedVariable] = [None] * len(tables)  # type: ignore
        _vars = out.ravel()._vars
        for i in range(len(tables)):
            ret_vars[i] = _vars[i].lookup(tables[i])
        out = FixedVariableArray(np.array(ret_vars).reshape(out_shape), solver_options=out.solver_options)
        out = np.sum(out, axis=-2)  # type: ignore
        return out


class ReplayConvTable(ReplayDenseTable):
    handles = (QConvT2D, QConvT1D, QConvTBase)

    def call(self, inputs: FixedVariableArray):
        op: QConvTBase = self.op

        if op.rank == 1:
            inputs = inputs[:, None]

        inputs = symbolic_extract_patches(inputs, **op.im2col_params)

        if op.rank == 1:
            inputs = inputs[:, 0]

        return super().call(inputs)


__all__ = ['ReplayDenseTable', 'ReplayConvTable']



from collections.abc import Callable, Sequence

import numpy as np
from keras import ops
from keras.callbacks import Callback
from keras.models import Model


class BetaScheduler(Callback):
    """Schedule the beta value of the Q Layers.

    Parameters
    ----------
    beta_fn : Callable[[int], float]
        A function that takes the current epoch and returns the beta value.
    """

    def __init__(self, beta_fn: Callable[[int], float]):
        self.beta_fn = beta_fn

    def on_epoch_begin(self, epoch, logs=None):
        assert isinstance(self.model, Model)

        beta = self.beta_fn(epoch)
        for layer in self.model._flatten_layers():
            if getattr(layer, '_beta', None) is not None:
                layer._beta.assign(ops.convert_to_tensor(beta, dtype=layer._beta.dtype))

    def on_epoch_end(self, epoch, logs=None):
        assert isinstance(logs, dict)
        logs['beta'] = self.beta_fn(epoch)


class PieceWiseSchedule:
    """Get interpolated schedule from key points.

    Parameters
    ----------
    intervals : sequence of tuple[epoch:int, beta:float, interp:str]
        The key points of the schedule. Each tuple contains the starting epoch, beta, and interpolation for the interval.

        epoch: the epoch number
        beta: the beta value at that epoch
        interp: the interpolation type in the interval after that epoch, one of 'linear', 'log', 'constant'. After the last epoch defined in the intervals, the beta value will always be constant disregarding the interpolation type.

        Example: `[(0, 0, 'linear'), (10, 1e-5, 'log'), (20, 1e-3, 'constant')]` will start with beta=0, then increase to 1e-5 in 10 epochs linearly, and increase to 1e-3 in another 10 epochs logarithmically. beta will stay at 1e-3 after 20 epochs.
    """

    def __init__(self, intervals: Sequence[tuple[int, float, str]]):
        intervals = sorted(intervals, key=lambda v: v[0])
        epochs = [v[0] for v in intervals]
        betas = [v[1] for v in intervals]
        interpolations = [v[2] for v in intervals]
        assert all(interp in ('linear', 'log', 'constant') for interp in interpolations)

        self.epochs = epochs
        self.betas = betas
        self.interpolations = interpolations

    def __call__(self, epoch):
        idx0 = np.searchsorted(self.epochs, epoch, side='right') - 1
        idx1 = idx0 + 1
        idx0 = max(0, min(idx0, len(self.epochs) - 1))
        idx1 = max(0, min(idx1, len(self.epochs) - 1))
        beta0, beta1 = self.betas[idx0], self.betas[idx1]
        epoch0, epoch1 = self.epochs[idx0], self.epochs[idx1]
        interp = self.interpolations[idx0]

        eps = 1e-9
        match interp:
            case 'linear':
                beta = beta0 + (beta1 - beta0) * (epoch - epoch0) / (epoch1 - epoch0 + eps)
            case 'log':
                beta = beta0 * (beta1 / beta0) ** ((epoch - epoch0) / (epoch1 - epoch0 + eps))
            case 'constant':
                beta = beta0
            case _:
                raise ValueError(f'Invalid interpolation type: {interp}')
        return float(beta)



from collections.abc import Callable
from math import prod
from typing import TypedDict

import keras
from keras import ops
from keras.layers import Layer
from keras.saving import register_keras_serializable
from keras.src.ops.operation_utils import compute_conv_output_shape

from ...quantizer import QuantizerConfig
from .dense import QDenseT


class im2col_params_t(TypedDict):
    size: tuple[int, int]
    strides: tuple[int, int]
    dilation_rate: tuple[int, int]
    padding: str
    data_format: str


class QConvTBase(QDenseT):
    def __init__(
        self,
        filters: int,
        kernel_size: int | tuple[int, ...],
        rank: int,
        n_hl: int = 1,
        d_hl: int = 8,
        strides: int | tuple[int, ...] = 1,
        padding='valid',
        data_format=None,
        dilation_rate: int | tuple[int, ...] = 1,
        groups: int = 1,
        activation: Callable | None | str = None,
        subnn_activation: str | Callable | Layer | None = None,
        use_bias=True,
        activity_regularizer=None,
        toq_conf: QuantizerConfig | None = None,
        parallelization_factor: int = -1,
        **kwargs,
    ):
        super().__init__(
            n_out=filters,
            n_hl=n_hl,
            d_hl=d_hl,
            activation=activation,
            subnn_activation=subnn_activation,
            toq_conf=toq_conf,
            parallelization_factor=parallelization_factor,
            use_bias=use_bias,
            **kwargs,
        )

        def _normalize(x):
            if isinstance(x, int):
                return (x,) * rank
            assert len(x) == rank, f'Expected {rank}-D tuple/list, got {x}'
            return tuple(x)

        self.rank = rank
        self.strides = _normalize(strides)
        self.padding = padding
        self.data_format = data_format or 'channels_last'
        self.dilation_rate = _normalize(dilation_rate)
        self.groups = groups
        self.kernel_size = _normalize(kernel_size)
        self.activity_regularizer = keras.regularizers.get(activity_regularizer)

        assert groups == 1, 'Table-based grouped convolutions are not supported yet.'

    def call(self, x, training=None):
        if self.rank == 1:
            x = x[:, :, None]
        x = ops.image.extract_patches(x, **self.im2col_params)  # type: ignore
        if self.rank == 1:
            x = x[:, :, 0]
        return super().call(x, training=training)

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        output_shape = compute_conv_output_shape(
            input_shape,
            filters=self.n_out,
            kernel_size=self.kernel_size,
            strides=self.strides,  # type: ignore
            padding=self.padding,
            data_format=self.data_format,
            dilation_rate=self.dilation_rate,  # type: ignore
        )
        ch_in = input_shape[-1]
        n_in = ch_in * prod(self.kernel_size)
        _dense_in_shape = output_shape[:-1] + (n_in,)
        self.output_shape = (1, *output_shape[1:])

        self._build_im2col_params()

        super().build(_dense_in_shape)

    def _build_im2col_params(self):
        if self.rank == 2:
            self.im2col_params: im2col_params_t = {  # type: ignore
                'size': self.kernel_size,
                'strides': self.strides,
                'dilation_rate': self.dilation_rate,
                'padding': self.padding,
                'data_format': self.data_format,
            }
        elif self.rank == 1:
            if self.data_format == 'channels_last':
                size = self.kernel_size + (1,)
                strides = self.strides + (1,)
                dilation_rate = self.dilation_rate + (1,)
            else:
                size = (1,) + self.kernel_size
                strides = (1,) + self.strides
                dilation_rate = (1,) + self.dilation_rate

            self.im2col_params: im2col_params_t = {  # type: ignore
                'size': size,
                'strides': strides,
                'dilation_rate': dilation_rate,
                'padding': self.padding,
                'data_format': self.data_format,
            }
        else:
            raise ValueError('Only 1D and 2D convolutions are supported.')

    def _compute_ebops(self, shape: tuple[int, ...]):
        q_shape = self.output_shape[:-1] + (self.n_in,)
        return super()._compute_ebops(q_shape)

    @property
    def toq(self):
        return self._toq

    def get_config(self):
        config = super().get_config()
        config['filters'] = config.pop('n_out')
        new_config = {
            'kernel_size': self.kernel_size,
            'strides': self.strides,
            'padding': self.padding,
            'data_format': self.data_format,
            'dilation_rate': self.dilation_rate,
            'groups': self.groups,
        }
        config.update(new_config)
        return config


@register_keras_serializable(package='hgq2')
class QConvT1D(QConvTBase):
    def __init__(
        self,
        filters: int,
        kernel_size: int,
        n_hl: int = 1,
        d_hl: int = 8,
        strides: int | tuple[int] = 1,
        padding='valid',
        data_format=None,
        dilation_rate: int | tuple[int] = 1,
        groups: int = 1,
        activation: Callable | None | str = None,
        subnn_activation: str | Callable | Layer | None = None,
        use_bias=True,
        activity_regularizer=None,
        toq_conf: QuantizerConfig | None = None,
        parallelization_factor: int = -1,
        **kwargs,
    ):
        super().__init__(
            filters=filters,
            kernel_size=kernel_size,
            rank=1,
            n_hl=n_hl,
            d_hl=d_hl,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            subnn_activation=subnn_activation,
            use_bias=use_bias,
            activity_regularizer=activity_regularizer,
            toq_conf=toq_conf,
            parallelization_factor=parallelization_factor,
            **kwargs,
        )


@register_keras_serializable(package='hgq2')
class QConvT2D(QConvTBase):
    def __init__(
        self,
        filters: int,
        kernel_size: int | tuple[int, int],
        n_hl: int = 1,
        d_hl: int = 8,
        strides: int | tuple[int, int] = 1,
        padding='valid',
        data_format=None,
        dilation_rate: int | tuple[int, int] = 1,
        groups: int = 1,
        activation: Callable | None | str = None,
        subnn_activation: str | Callable | Layer | None = None,
        use_bias=True,
        activity_regularizer=None,
        toq_conf: QuantizerConfig | None = None,
        parallelization_factor: int = -1,
        **kwargs,
    ):
        super().__init__(
            filters=filters,
            kernel_size=kernel_size,
            rank=2,
            n_hl=n_hl,
            d_hl=d_hl,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            subnn_activation=subnn_activation,
            use_bias=use_bias,
            activity_regularizer=activity_regularizer,
            toq_conf=toq_conf,
            parallelization_factor=parallelization_factor,
            **kwargs,
        )



import keras
import numpy as np
import pytest
from keras import ops

from hgq.layers.attn import QLinformerAttention, QMultiHeadAttention, QSALTAttention
from hgq.quantizer.internal import FixedPointQuantizerKBI, FixedPointQuantizerKIF

from .base import LayerTestBase


class TestMultiHeadAttention(LayerTestBase):
    layer_cls = QMultiHeadAttention

    @pytest.fixture(params=[1, 2])  # Test different head counts
    def num_heads(self, request):
        return request.param

    @pytest.fixture(params=[8])  # Test different key dimensions
    def key_dim(self, request):
        return request.param

    @pytest.fixture(params=['qkv', 'kv', 'none'])  # Test different fusion modes
    def fuse(self, request):
        # qkv and kv fused projection only makes difference in da4ml
        return request.param

    @pytest.fixture
    def input_shapes(self, fuse: str):
        # Seq..., dim
        if fuse == 'none':
            return (2, 4, 8), (2, 4, 9), (2, 4, 10)
        elif fuse == 'qkv':
            return ((9, 8),)
        elif fuse == 'kv':
            return (2, 4, 3), (2, 4, 5)
        raise ValueError(f'Invalid fusion mode: {fuse}')

    @pytest.fixture
    def value_dim(self, key_dim):
        return key_dim if self.fuse in ('qkv', 'kv') else key_dim + 1

    @pytest.fixture
    def layer_kwargs(self, num_heads, key_dim, fuse):
        return {'num_heads': num_heads, 'key_dim': key_dim, 'fuse': fuse}

    @pytest.fixture(params=[True])
    def use_parallel_io(self, request) -> bool:
        return request.param

    @pytest.fixture
    def call_kwargs(self, input_shapes):
        if len(input_shapes) == 1:
            return {'use_causal_mask': True}
        return {}

    @pytest.fixture
    def model(self, layer, input_shapes, call_kwargs, use_parallel_io):
        """Create test model with the layer"""
        if isinstance(input_shapes[0], int):
            input_shapes = (input_shapes,)
        inputs = [keras.layers.Input(shape=shape) for shape in input_shapes]

        outputs = layer(*inputs, **call_kwargs)
        model = keras.Model(inputs, outputs)

        self.perturbe_bw(use_parallel_io, model)
        return model

    def perturbe_bw(self, use_parallel_io, model):
        if use_parallel_io:
            for _layer in model._flatten_layers(False):
                if isinstance(_layer, FixedPointQuantizerKBI):
                    b = np.random.randint(4, 8, _layer._b.shape)
                    i = ops.convert_to_numpy(ops.stop_gradient(_layer.i))
                    b = np.minimum(b, 12 - i)  # type: ignore
                    if np.all(b == 0):
                        b.ravel()[0] = 1
                    _layer._b.assign(ops.array(b))
                if isinstance(_layer, FixedPointQuantizerKIF):
                    f = np.random.randint(2, 8, _layer._f.shape)
                    i = ops.convert_to_numpy(ops.stop_gradient(_layer.i))
                    f = np.minimum(f, 12 - i)  # type: ignore
                    if np.all(i + f == 0):
                        f.ravel()[0] = 1
                    _layer._f.assign(ops.array(f))
        for _layer in model._flatten_layers(False):
            # Randomize activation values
            if hasattr(_layer, 'bias') and isinstance(_layer.bias, keras.Variable):
                bias = np.random.randn(*_layer.bias.shape)
                _layer.bias.assign(ops.array(bias))

    @pytest.fixture
    def input_data(self, input_shapes, N: int = 5000):
        eps = 2.0**-8
        return tuple(
            np.round((np.random.randn(N, *shape).astype(np.float32) * 3).clip(-8, 8 - eps) * 256) / 256 for shape in input_shapes
        )

    def assert_equal(self, keras_output, hls_output):
        return np.testing.assert_allclose(keras_output, hls_output, atol=1e-6)

    def test_hls4ml_conversion(  # type: ignore
        self, model: keras.Model, input_data, temp_directory: str, use_parallel_io: bool, q_type: str, call_kwargs
    ):
        if call_kwargs.get('use_causal_mask', False):
            pytest.skip('Causal mask not supported in hls4ml conversion')
        super().test_hls4ml_conversion(model, input_data, temp_directory, use_parallel_io, q_type)


class TestLinformerAttention(TestMultiHeadAttention):
    layer_cls = QLinformerAttention

    @pytest.fixture
    def layer_kwargs(self, num_heads, key_dim, fuse, input_shapes):
        lin_kv_proj_dim = [2 for _ in range(len(input_shapes[0]) - 1)]
        return {'num_heads': num_heads, 'key_dim': key_dim, 'fuse': fuse, 'lin_kv_proj_dim': lin_kv_proj_dim}

    @pytest.fixture(params=['none'])  # Test different fusion modes
    def fuse(self, request):
        return request.param

    @pytest.fixture
    def call_kwargs(self, input_shapes):
        return {}


class TestSALTAttention(TestMultiHeadAttention):
    layer_cls = QSALTAttention
    hls4ml_not_supported = True

    @pytest.fixture(params=['none'])  # Test different fusion modes
    def fuse(self, request):
        return request.param

    @pytest.fixture(params=[(True, True, True), (True, False, False)])
    def cluster_and_share(self, request):
        return request.param

    @pytest.fixture
    def input_shapes(self):
        return ((9, 12),)

    @pytest.fixture
    def layer_kwargs(self, num_heads, key_dim, fuse, cluster_and_share):
        cluster_k_proj, cluster_v_proj, share_kv_proj = cluster_and_share
        lin_kv_proj_dim = 4
        return {
            'num_heads': num_heads,
            'key_dim': key_dim,
            'fuse': fuse,
            'cluster_k_proj': cluster_k_proj,
            'cluster_v_proj': cluster_v_proj,
            'share_kv_proj': share_kv_proj,
            'lin_kv_proj_dim': lin_kv_proj_dim,
        }

    @pytest.fixture
    def call_kwargs(self, input_shapes):
        return {}



import typing
from collections.abc import Sequence
from typing import Any

import keras
import numpy as np
from da4ml.trace import FixedVariable, FixedVariableArray
from da4ml.trace.ops import quantize, relu
from keras.ops import convert_to_numpy

import hgq
from hgq.layers.core.base import MultipleQuantizers, Quantizer
from hgq.quantizer.internal import FixedPointQuantizerBase


def to_np_arr(x: Any) -> np.ndarray:
    return np.asarray(convert_to_numpy(x))


def mirror_quantizer(q: Quantizer, v: FixedVariableArray) -> FixedVariableArray:
    if q.scaler is not None:
        v = v * (1.0 / q.scaler)
    q_internal: FixedPointQuantizerBase = q.quantizer
    kk, ki, kf = q_internal.kif
    shape = (1,) + v.shape
    kk = q_internal.bw_mapper.bw_to_x(kk, shape)
    ki = q_internal.bw_mapper.bw_to_x(ki, shape)
    kf = q_internal.bw_mapper.bw_to_x(kf, shape)
    k, i, f = (to_np_arr(x).astype(np.int8)[0] for x in (kk, ki, kf))
    round_mode, overflow_mode = q_internal.round_mode, q_internal.overflow_mode
    rq = quantize(v, k, i, f, overflow_mode=overflow_mode, round_mode=round_mode)
    if q.affine:
        rq = rq * q.affine[0] + q.affine[1]
    return rq


_registry: dict[type, 'type[ReplayOperationBase]'] = {}


class HandlerRegMeta(type):
    """Metaclass for automatic registration of handler classes."""

    def __new__(mcs, name: str, bases: tuple[type, ...], namespace: dict[str, typing.Any]):
        cls = super().__new__(mcs, name, bases, namespace)
        if name == 'ReplayOperationBase':
            return cls

        handles: type | tuple[type, ...] = namespace['handles']
        if not isinstance(handles, tuple):
            handles = (handles,)

        for handle in handles:
            _registry[handle] = cls  # type: ignore
        return cls


class ReplayOperationBase(metaclass=HandlerRegMeta):
    handles: tuple[type, ...] = ()
    __activation_handled__ = False
    __input_quantizer_handled__ = False
    __output_quantizer_handled__ = False

    def __init__(self, layer: 'keras.Operation'):
        assert isinstance(layer, self.handles)
        self.op: Any = layer

    def call(self, *args, **kwargs) -> tuple[FixedVariableArray, ...] | FixedVariableArray: ...

    def __call__(self, *args, **kwargs) -> tuple[FixedVariableArray, ...]:
        assert all(not isinstance(a, FixedVariableArray) for a in kwargs.values())

        if not isinstance(self.op, hgq.layers.QLayerBase):
            r = self.call(*args, **kwargs)
            return r if isinstance(r, tuple) else (r,)

        layer: hgq.layers.QLayerBase = self.op
        assert kwargs.pop('training', False) is False, 'Training mode is not supported in mirror operation'

        if not self.__input_quantizer_handled__:
            assert len(args) == 1
            inputs = args[0]

            if layer.enable_iq:
                if isinstance(inputs, Sequence):
                    assert isinstance(layer.iq, MultipleQuantizers)
                    inputs = tuple(mirror_quantizer(q, v) for q, v in zip(layer.iq.quantizers, inputs))
                else:
                    assert isinstance(layer.iq, Quantizer), f'Expected iq to be a Quantizer, got {type(layer.iq)}'
                    inputs = mirror_quantizer(layer.iq, inputs)

            outputs = self.call(inputs, **kwargs)
        else:
            outputs = self.call(*args, **kwargs)
        if isinstance(outputs, FixedVariable):
            outputs = FixedVariableArray(np.array([outputs]))

        if not self.__activation_handled__:
            activation = getattr(layer, 'activation', keras.activations.linear)
            if activation is not keras.activations.linear:
                if activation is keras.activations.relu:
                    if isinstance(outputs, tuple):
                        assert len(outputs) == 1, 'ReLU activation is expected to have a single output'
                        outputs = (relu(outputs[0]),)
                    else:
                        outputs = relu(outputs)
                else:
                    raise NotImplementedError(
                        f'Activation {activation} is not allowed in activation= field for common layers.'
                        ' Use dedicated QUnaryFunctionLUT layer instead.'
                    )

        if layer.enable_oq and not self.__output_quantizer_handled__:
            if isinstance(outputs, tuple):
                assert isinstance(layer.oq, MultipleQuantizers)
                outputs = tuple(mirror_quantizer(q, v) for q, v in zip(layer.oq.quantizers, outputs))
            else:
                assert isinstance(layer.oq, Quantizer)
                outputs = mirror_quantizer(layer.oq, outputs)

        if isinstance(outputs, (FixedVariableArray, np.ndarray)):
            outputs = (outputs,)

        return outputs


class ReplayQuantizer(ReplayOperationBase):
    handles = (Quantizer,)

    def __init__(self, op: 'Quantizer'):
        super().__init__(op)
        assert isinstance(op.quantizer, FixedPointQuantizerBase)

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        return mirror_quantizer(self.op, inputs)



from collections.abc import Sequence

import keras
import numpy as np
from keras import ops
from keras.utils import PyDataset
from numpy.typing import ArrayLike, NDArray
from tqdm import tqdm

from .dataset import Dataset


def _reset_minmax(layer: keras.Layer):
    if hasattr(layer, '_i_decay_speed') and layer.trainable:
        # WRAP-like overflow mode
        shape, dtype = layer._i.shape, layer._i.dtype
        layer._i.assign(keras.ops.full(shape, -1e9, dtype=dtype))
        shape, dtype = layer._k.shape, layer._k.dtype
        layer._k.assign(keras.ops.zeros(shape, dtype=dtype))
    for sublayer in layer._layers:
        _reset_minmax(sublayer)


class TrainingFlagWrapper:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other

    def __bool__(self):
        return self.value is True


def trace_minmax(
    model: keras.Model,
    data: ArrayLike | Sequence[ArrayLike] | PyDataset,
    reset=True,
    batch_size=1024,
    verbose: int | bool = 0,
    return_results=False,
) -> int | NDArray | tuple[NDArray, ...]:
    """With a calibration dataset, find the necessary integer bits required for the quantizers in a model.
    Only needed if `WRAP` overflow mode is used anywhere for the activation quantizers.

    Parameters
    ==========
    model: keras.Model
        The model to trace.
    data: ArrayLike or Sequence[ArrayLike] or PyDataset
        The calibration dataset.
    reset: bool, default True
        Whether to reset the min/max values before tracing. Set to False if you want to continue tracing from previous values.
    batch_size: int, default 1024
        The batch size to use for tracing.
    verbose: int or bool, default 0
        If not o or False, print the EBOPs for each layer after tracing.
        If > 1 or True, show a progress bar during tracing.
    return_results: bool, default False
        If True, return the model outputs on the calibration dataset. If False, return the total EBOPs.
    """
    n_outputs = len(model.outputs)

    if not isinstance(data, PyDataset):
        data = Dataset(data, batch_size=batch_size, device='none', shuffle=False)

    if reset:
        _reset_minmax(model)
    record: dict[str, int] = {}

    results = []
    use_pbar = verbose is True or verbose > 1
    n_batch = len(data)  # type: ignore
    n_outputs = len(model.outputs)

    with tqdm(total=n_batch, leave=False, disable=not use_pbar, desc='Tracing min/max') as pbar:
        for i in range(n_batch):
            r = model(data[i][0], training=TrainingFlagWrapper('tracing'))
            if return_results:
                results.append(ops.convert_to_numpy(r))
            pbar.update(1)

    record = {}
    for layer in model.layers:
        if getattr(layer, 'enable_ebops', False):
            record[layer.name] = int(layer.ebops)  # type: ignore

    if verbose:
        width = max(max(map(len, record.keys())), 5)
        for k, v in record.items():
            print(f'{k:{width}}: {v}')
        print(f'Total: {sum(record.values())}')

    if return_results:
        if n_outputs == 1:
            return np.concatenate([r for r in results])
        return tuple(np.concatenate([r[i] for r in results]) for i in range(n_outputs))
    else:
        return int(sum(record.values()))



from math import ceil
from typing import Literal

from keras import ops

from ...quantizer.config import QuantizerConfig
from ...utils.misc import gather_vars_to_kwargs
from ..conv import QConv2D
from ..core.einsum_dense import QEinsumDense
from .linformer import QMultiHeadAttention


class QSALTAttention(QMultiHeadAttention):
    """Quantized version of SALT attention from "Spatially Aware Linear Transformer (SALT) for Particle Jet Tagging"
    (https://arxiv.org/abs/2510.23641). Adds a conv2d operation on the computed attnmask before applying it to the value tensor.
    """

    def __init__(
        self,
        num_heads,
        lin_kv_proj_dim,
        key_dim,
        value_dim=None,
        dropout=0.0,
        use_bias=True,
        output_shape=None,
        attention_axes=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        seed=None,
        fuse: Literal['none'] = 'none',
        qkvo_iq_conf: QuantizerConfig | None = None,
        qkvo_kq_conf: QuantizerConfig | None = None,
        qkvo_bq_conf: QuantizerConfig | None = None,
        qkvo_oq_conf: QuantizerConfig | None = None,
        softmax_iq_conf: QuantizerConfig | None = None,
        softmax_exp_iq_conf: QuantizerConfig | None = None,
        softmax_exp_oq_conf: QuantizerConfig | None = None,
        softmax_inv_iq_conf: QuantizerConfig | None = None,
        softmax_inv_oq_conf: QuantizerConfig | None = None,
        softmax_oq_conf: QuantizerConfig | None = None,
        stable_softmax=True,
        softmax_allow_heterogeneous_table: bool = False,
        parallelization_factor=-1,
        share_kv_proj=False,
        cluster_k_proj=False,
        cluster_v_proj=False,
        conv_size=3,
        separate_conv=False,
        **kwargs,
    ):
        if fuse != 'none':
            raise ValueError(f'Only fuse="none" can be used for QLinformerAttention, but got fuse="{fuse}".')
        kwargs = gather_vars_to_kwargs('self|lin_kv_proj_dim|share_kv_proj|conv_size|cluster_k_proj|cluster_v_proj|separate_conv')
        self._kv_proj_dim = lin_kv_proj_dim if isinstance(lin_kv_proj_dim, int) else tuple(lin_kv_proj_dim)
        self.share_kv_proj = share_kv_proj
        self.conv_size = conv_size
        self.cluster_k_proj = cluster_k_proj
        self.cluster_v_proj = cluster_v_proj
        self.separate_conv = separate_conv
        super().__init__(**kwargs)

    def build(self, query_shape, value_shape=None, key_shape=None):
        value_shape = value_shape or query_shape
        key_shape = key_shape or value_shape
        key_rank = len(key_shape)
        value_rank = len(value_shape)
        assert key_rank == value_rank == len(query_shape) == 3, (
            f'Only rank 3 inputs are supported for SALTAttn, but got query ({query_shape}), key ({key_shape}), and value ({value_shape}).'
        )
        assert key_rank == value_rank, (
            f'Key and value must have the same rank, but got key shape {key_shape} and value shape {value_shape}.'
        )
        if self.share_kv_proj:
            assert key_shape == value_shape, (
                f'When share_kv_proj is True, k and v must have the same shape. Got key ({key_shape}) and value ({value_shape}).'
            )
            assert self.cluster_k_proj == self.cluster_v_proj, (
                f'When share_kv_proj is True, cluster_k_proj and cluster_v_proj must be the same. Got cluster_k_proj={self.cluster_k_proj} and cluster_v_proj={self.cluster_v_proj}.'
            )

        if self._attention_axes is None:
            self._attention_axes = tuple(range(1, value_rank - 1))
        elif not isinstance(self._attention_axes, tuple):
            self._attention_axes = tuple(self._attention_axes)

        _value_shape_proj, _key_shape_proj = list(value_shape), list(key_shape)
        _value_shape_proj[1] = self._kv_proj_dim
        _key_shape_proj[1] = self._kv_proj_dim
        self._key_shape_proj = tuple(_key_shape_proj)
        self._value_shape_proj = tuple(_value_shape_proj)

        eq_lin_k_proj = 'bmnd,mn->bmd' if self.cluster_k_proj else 'bnd,nm->bmd'  # x is 1, rm'rd here
        eq_lin_v_proj = 'bmnd,mn->bmd' if self.cluster_v_proj else 'bnd,nm->bmd'

        # padding config
        self.n_k_pad = 0
        self.n_v_pad = 0
        if self.cluster_k_proj:
            self.n_k_pad = ceil(key_shape[1] / self._kv_proj_dim) * self._kv_proj_dim - key_shape[1]
            cluster_length = (key_shape[1] + self.n_k_pad) // self._kv_proj_dim
            key_shape = (key_shape[0], self._kv_proj_dim, cluster_length, key_shape[2])
            self._k_reshape_to = (-1,) + key_shape[1:]
        if self.cluster_v_proj:
            self.n_v_pad = ceil(value_shape[1] / self._kv_proj_dim) * self._kv_proj_dim - value_shape[1]
            cluster_length = (value_shape[1] + self.n_v_pad) // self._kv_proj_dim
            value_shape = (value_shape[0], self._kv_proj_dim, cluster_length, value_shape[2])
            self._v_reshape_to = (-1,) + value_shape[1:]

        self._lin_k_proj = QEinsumDense(
            eq_lin_k_proj, self._key_shape_proj[1:], bias_axes=None, **self._get_common_kwargs_for_sublayer()
        )
        self._lin_k_proj.build(key_shape)

        if not self.share_kv_proj:
            self._lin_v_proj = QEinsumDense(
                eq_lin_v_proj, self._value_shape_proj[1:], bias_axes=None, **self._get_common_kwargs_for_sublayer()
            )
            self._lin_v_proj.build(value_shape)
        else:
            self._lin_v_proj = self._lin_k_proj

        attn_score_shape = (query_shape[0], self.num_heads, query_shape[1], self._kv_proj_dim)
        print(attn_score_shape)
        if self.conv_size > 0:
            self.conv = QConv2D(
                filters=attn_score_shape[1],
                kernel_size=self.conv_size,
                padding='same',
                groups=self.num_heads if self.separate_conv else 1,
                **self._get_common_kwargs_for_sublayer(),
                data_format='channels_first',
            )
            self.conv.build(attn_score_shape)

        super().build(query_shape, self._value_shape_proj, key_shape=self._key_shape_proj)

    def _masked_softmax(self, attention_scores, attention_mask=None):
        if self.conv_size > 0:
            attention_scores = self.conv(attention_scores)
        attention_scores = super()._masked_softmax(attention_scores, attention_mask)
        return attention_scores

    def call(
        self,
        query,
        value=None,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        training=None,
        use_causal_mask=False,
    ):
        assert use_causal_mask is False, 'Causal mask is not supported in QLinformerAttention.'
        value = value if value is not None else query
        key = key if key is not None else value

        if self.cluster_k_proj:
            key = ops.pad(key, [[0, 0], [0, self.n_k_pad], [0, 0]])
            key = ops.reshape(key, self._k_reshape_to)

        if self.n_v_pad > 0:
            value = ops.pad(value, [[0, 0], [0, self.n_v_pad], [0, 0]])
            value = ops.reshape(value, self._v_reshape_to)

        key = self._lin_k_proj(key, training=training)
        value = self._lin_v_proj(value, training=training)

        return super().call(
            query,
            value,
            key,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            return_attention_scores=return_attention_scores,
            training=training,
        )

    @property
    def ebops(self):
        if self._ebops is None:
            return ops.cast(0, 'uint32')
        ebops = sum(
            (  # type: ignore
                self._lin_k_proj.ebops,
                self._lin_v_proj.ebops,
                ops.convert_to_tensor(self._ebops),  # type: ignore
            )
        )
        if self.conv_size > 0:
            ebops += self.conv.ebops  # type: ignore
        return ebops  # type: ignore

    def _compute_ebops(self, query_shape, value_shape=None, key_shape=None):
        return super()._compute_ebops(query_shape, self._value_shape_proj, key_shape=self._key_shape_proj)

    def get_config(self):
        config = super().get_config()
        config['lin_kv_proj_dim'] = self._kv_proj_dim
        config['share_kv_proj'] = self.share_kv_proj
        config['conv_size'] = self.conv_size
        config['cluster_k_proj'] = self.cluster_k_proj
        config['cluster_v_proj'] = self.cluster_v_proj
        config['separate_conv'] = self.separate_conv
        return config



import numpy as np
import pytest
from keras import ops
from keras.initializers import RandomNormal, RandomUniform

from hgq.config import QuantizerConfigScope
from hgq.layers import QBatchNormalization

from .base import LayerTestBase


class TestBatchNorm(LayerTestBase):
    layer_cls = QBatchNormalization

    @pytest.fixture(params=[2])
    def axis(self, request) -> int:
        return request.param

    @pytest.fixture(params=[True])
    def scale(self, request) -> bool:
        return request.param

    @pytest.fixture(params=[True])
    def center(self, request) -> bool:
        return request.param

    @pytest.fixture
    def layer_kwargs(self, axis, scale, center):
        return {
            'axis': axis,
            'scale': scale,
            'center': center,
            'momentum': 0.99,
            'synchronized': False,
            'moving_mean_initializer': RandomNormal(stddev=1.0),
            'moving_variance_initializer': RandomUniform(minval=0.0, maxval=8.0),
        }

    @pytest.fixture
    def input_shapes(self):
        return (6, 8)

    def test_behavior(self, input_data, layer_kwargs):
        with QuantizerConfigScope(default_q_type='dummy'):
            bn = QBatchNormalization(**{**layer_kwargs, 'momentum': 0.0})  # type: ignore

        input_data = ops.convert_to_tensor(input_data, dtype=bn.dtype)  # type: ignore
        hgq_output = bn(input_data, training=True)
        hgq_output_test = bn(input_data, training=False)
        mean, var = ops.moments(input_data, axes=layer_kwargs['axis'], keepdims=True)  # type: ignore
        ref_output = (input_data - mean) / ops.sqrt(var + bn.epsilon)  # type: ignore

        hgq_output_np: np.ndarray = ops.convert_to_numpy(hgq_output)  # type: ignore
        ref_output_np: np.ndarray = ops.convert_to_numpy(ref_output)  # type: ignore
        hgq_output_test_np: np.ndarray = ops.convert_to_numpy(hgq_output_test)  # type: ignore

        np.allclose(hgq_output_np, ref_output_np, atol=1e-6)
        np.allclose(hgq_output_test_np, ref_output_np)



import keras
from keras import ops
from keras.callbacks import EarlyStopping

from ...layers.core.base import QLayerBase
from ..misc import gather_vars_to_kwargs


class EarlyStoppingWithEbopsThres(EarlyStopping):
    """Vanilla Keras EarlyStopping but only after a given EBOPs threshold is met.

    This callback stops training when:
    - EBOPs is lower than a given threshold, and
    - monitored metric has stopped improving

    Assuming the goal of a training is to minimize the loss. With this, the
    metric to be monitored would be `'loss'`, and mode would be `'min'`. A
    `model.fit()` training loop will check at end of every epoch whether
    the loss is no longer decreasing, considering the `min_delta` and
    `patience` if applicable. Once it's found no longer decreasing,
    `model.stop_training` is marked True and the training terminates.

    The quantity to be monitored needs to be available in `logs` dict.
    To make it so, pass the loss or metrics at `model.compile()`.

    Parameters
    ----------
    ebops_threshold : float
        The target EBOps value. This callback will not stop the
        training until the model's EBOPs is at or below this value.
    monitor : str, default "val_loss"
        Quantity to be monitored.
    min_delta : float, default 0
        Minimum change in the monitored quantity to qualify as an
        improvement, i.e. an absolute change of less than min_delta, will
        count as no improvement.
    patience : int, default 0
        Number of epochs with no improvement after which training will
        be stopped.
    verbose : int, default 0
        Verbosity mode, 0 or 1. Mode 0 is silent, and mode 1 displays
        messages when the callback takes an action.
    mode : {"auto", "min", "max"}, default "auto"
        In `min` mode, training will stop when the quantity monitored has
        stopped decreasing; in `"max"` mode it will stop when the quantity
        monitored has stopped increasing; in `"auto"` mode, the direction
        is automatically inferred from the name of the monitored quantity.
    baseline : float, optional
        Baseline value for the monitored quantity. If not `None`,
        training will stop if the model doesn't show improvement over the
        baseline.
    restore_best_weights : bool, default False
        Whether to restore model weights from the epoch with the best value
        of the monitored quantity. If `False`, the model weights obtained at
        the last step of training are used. An epoch will be restored
        regardless of the performance relative to the `baseline`. If no epoch
        improves on `baseline`, training will run for `patience` epochs and
        restore weights from the best epoch in that set.
    start_from_epoch : int, default 0
        Number of epochs to wait before starting to monitor improvement.
        This allows for a warm-up period in which no improvement is expected
        and thus training will not be stopped.
    """

    def __init__(
        self,
        ebops_threshold: float,
        monitor='val_loss',
        min_delta: float = 0,
        patience: int = 0,
        verbose: int = 0,
        mode='auto',
        baseline: float | None = None,
        restore_best_weights: bool = False,
        start_from_epoch: int = 0,
    ):
        self.ebops_threshold = ebops_threshold
        kwargs = gather_vars_to_kwargs('self|ebops_threshold')
        super().__init__(**kwargs)

    def on_epoch_end(self, epoch, logs=None):
        assert isinstance(self.model, keras.Model)
        stop_training = self.model.stop_training

        ebops = 0
        for layer in self.model._flatten_layers():
            if isinstance(layer, QLayerBase) and layer.enable_ebops:
                ebops += int(ops.convert_to_numpy(layer._ebops))  # type: ignore

        super().on_epoch_end(epoch, logs)
        if ebops > self.ebops_threshold:
            self.model.stop_training = stop_training



from collections.abc import Sequence

import keras
import numpy as np
from da4ml.trace import FixedVariableArray
from da4ml.trace.ops import einsum
from keras.src.ops.numpy import (
    Abs,
    Absolute,
    Add,
    Concatenate,
    Divide,
    Dot,
    Einsum,
    GetItem,
    Matmul,
    Max,
    Maximum,
    Min,
    Minimum,
    Moveaxis,
    Multiply,
    Ravel,
    Repeat,
    Reshape,
    Subtract,
    Sum,
    Transpose,
    TrueDivide,
)

from hgq.layers import (
    QAdd,
    QAveragePow2,
    QDot,
    QEinsum,
    QMaximum,
    QMeanPow2,
    QMinimum,
    QMultiply,
    QSubtract,
    QSum,
)

from ._base import ReplayOperationBase


class ReplayReshape(ReplayOperationBase):
    handles = (keras.layers.Reshape, keras.layers.Flatten, Reshape, Ravel)

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        if isinstance(self.op, (keras.layers.Flatten, Ravel)):
            return inputs.ravel()
        elif isinstance(self.op, keras.layers.Reshape):
            return inputs.reshape(self.op.target_shape)
        elif isinstance(self.op, Reshape):
            return inputs.reshape(self.op.newshape[1:])
        else:
            raise TypeError(f'Unsupported layer type: {type(self.op)}')


class ReplayMerge(ReplayOperationBase):
    handles = (keras.layers.Add, keras.layers.Concatenate, QAdd, QMultiply, QSubtract, QMaximum, QMinimum, QAveragePow2)

    def call(self, inputs: tuple[FixedVariableArray, ...]) -> FixedVariableArray:
        op = self.op
        name = op.__class__.__name__
        if name.startswith('Q'):
            name = name[1:]
        _inputs: FixedVariableArray = np.stack(np.broadcast_arrays(*inputs), axis=0)  # type: ignore
        match name:
            case 'Add':
                return np.sum(_inputs, axis=0)  # type: ignore
            case 'AveragePow2':
                return np.sum(_inputs, axis=0) * op._scale  # type: ignore
            case 'Subtract':
                assert len(_inputs) == 2, 'Subtract operation requires exactly two inputs'
                return _inputs[0] - _inputs[1]
            case 'Multiply':
                return np.prod(_inputs, axis=0)  # type: ignore
            case 'Maximum':
                return np.amax(_inputs, axis=0)  # type: ignore
            case 'Minimum':
                return np.amin(_inputs, axis=0)  # type: ignore
            case 'Concatenate':
                return np.concatenate(_inputs, axis=op.axis)  # type: ignore

            case _:
                raise TypeError(f'Unsupported layer type: {type(op)}')


class ReplayRepeatVector(ReplayOperationBase):
    handles = (keras.layers.RepeatVector,)

    def call(self, inputs: FixedVariableArray) -> FixedVariableArray:
        layer: keras.layers.RepeatVector = self.op
        if layer.n == 1:
            return inputs
        # return FixedVariableArray(np.repeat(inputs._vars, layer.n, axis=0), inputs.solver_options)
        return np.repeat(inputs[None], layer.n, axis=0)[0]  # type: ignore


class ReplayGetItem(ReplayOperationBase):
    handles = (GetItem,)

    def call(self, x: FixedVariableArray, key):
        if isinstance(key, list):
            key = tuple(key)
        return x[None][key][0]


class ReplayReduction(ReplayOperationBase):
    handles = (Sum, Max, Min)

    def call(self, x: FixedVariableArray, axis=None, keepdims=False):
        if isinstance(self.op, Sum):
            op = np.sum
        elif isinstance(self.op, Max):
            op = np.amax
        elif isinstance(self.op, Min):
            op = np.amin
        return op(x[None], axis=axis, keepdims=keepdims)[0]  # type: ignore


class ReplayQReduction(ReplayOperationBase):
    handles = (QSum, QMeanPow2)

    def call(self, x: FixedVariableArray):
        layer: QSum = self.op
        axes, scale, keepdims = layer.axes, layer.scale, layer.keepdims
        return np.sum(x[None], axis=axes, keepdims=keepdims)[0] * scale  # type: ignore


class ReplayArithmetic(ReplayOperationBase):
    handles = (Add, Subtract, Multiply, TrueDivide, Divide, Maximum, Minimum)

    def call(self, x1: FixedVariableArray, x2: FixedVariableArray):
        name = self.op.__class__.__name__
        match name:
            case 'Add':
                return x1 + x2
            case 'Subtract':
                return x1 - x2
            case 'Multiply':
                return x1 * x2
            case 'TrueDivide' | 'Divide':
                return x1 / x2
            case 'Maximum':
                return np.maximum(x1, x2)  # type: ignore
            case 'Minimum':
                return np.minimum(x1, x2)  # type: ignore
            case _:
                raise TypeError(f'Unsupported arithmetic operation: {type(self.op)}')


class ReplayConcatenate(ReplayOperationBase):
    handles = (Concatenate,)

    def call(self, xs: Sequence[FixedVariableArray]):
        axis = self.op.axis
        # return backend.numpy.concatenate(xs, axis=self.axis)
        # return FixedVariableArray(np.concatenate([x._vars[None] for x in xs], axis=axis)[0], xs[0].solver_options)
        return np.concatenate([x[None] for x in xs], axis=axis)[0]  # type: ignore


class ReplayRepeat(ReplayOperationBase):
    handles = (Repeat,)

    def call(self, x: FixedVariableArray):
        repeats, axis = self.op.repeats, self.op.axis
        # return FixedVariableArray(np.repeat(x._vars[None], repeats, axis=axis)[0], x.solver_options)
        return np.repeat(x[None], repeats, axis=axis)[0]  # type: ignore


class ReplayTranspose(ReplayOperationBase):
    handles = (Transpose,)

    def call(self, x: FixedVariableArray):
        axes = self.op.axes
        return np.transpose(x, axes)  # type: ignore


class ReplayMoveaxis(ReplayOperationBase):
    handles = (Moveaxis,)

    def call(self, x: FixedVariableArray):
        source, destination = self.op.source, self.op.destination
        return np.moveaxis(x[None], source, destination)[0]  # type: ignore


class ReplayNoOp(ReplayOperationBase):
    __noop_layers = []
    for k, v in keras.layers.__dict__.items():
        name = k.lower()
        if 'dropout' in name or 'random' in name or 'noise' in name:
            __noop_layers.append(v)

    handles = tuple(__noop_layers)

    def call(self, x: FixedVariableArray, training=False) -> FixedVariableArray:
        assert not training, 'Training mode is not supported in mirror operation'
        return x


class ReplayEinsum(ReplayOperationBase):
    handles = (QEinsum, Einsum, QDot, keras.layers.Dot)

    def call(self, *_inputs: tuple[FixedVariableArray, FixedVariableArray] | FixedVariableArray) -> FixedVariableArray:
        op = self.op
        inputs: tuple[FixedVariableArray, FixedVariableArray]
        if isinstance(_inputs[0], tuple):
            assert len(_inputs) == 1, 'Einsum with multiple input tuples is not supported'
            inputs = _inputs[0]
        else:
            inputs = _inputs  # type: ignore
        assert len(inputs) == 2, 'Only (Q)Einsum operations with exactly two inputs are supported'
        if isinstance(op, QEinsum):
            eq = op.equation
        elif isinstance(op, Einsum):
            eq = op.subscripts
        else:  # QDot/Dot
            dim0, dim1 = inputs[0].ndim + 1, inputs[1].ndim + 1
            letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'[0 : dim0 + dim1]
            sub0, sub1 = letters[:dim0], letters[dim0 : dim0 + dim1]
            axes = list(op.axes) if not isinstance(op.axes, int) else [op.axes, op.axes]
            idx0, idx1 = axes[0] if axes[0] >= 0 else axes[0] % dim0, axes[1] if axes[1] >= 0 else axes[1] % dim1
            sub1 = sub1[:idx1] + sub0[idx0] + sub1[idx1 + 1 :]
            sub_out = list(sub0 + sub1)
            sub_out.remove(sub0[idx0])
            sub_out.remove(sub0[idx0])
            sub_out = ''.join(sub_out)
            eq = f'{sub0},{sub1}->{sub_out}'
        return einsum(eq, inputs[0][None], inputs[1][None])[0]


class ReplayMatmul(ReplayOperationBase):
    handles = (Matmul, Dot)

    def call(self, x1: FixedVariableArray, x2: FixedVariableArray) -> FixedVariableArray:
        return x1 @ x2


class ReplayAbs(ReplayOperationBase):
    handles = (Absolute, Abs)

    def call(self, x: FixedVariableArray) -> FixedVariableArray:
        return np.abs(x)  # type: ignore



from math import log10

from keras import ops
from keras.callbacks import Callback
from keras.models import Model


class PID:
    def __init__(self, p: float, i: float, d: float, neg: bool):
        self.p = p
        self.i = i
        self.d = d
        self.integral = 0.0
        self.prev_error = 0.0
        self.neg = neg

    def update_gains(self, p: float, i: float, d: float):
        self.p = p
        self.i = i
        self.d = d

    def update_gains_KcTiTd(self, Kp: float, Ti: float, Td: float):
        self.p = Kp
        self.i = Kp / Ti
        self.d = Kp * Td

    def __call__(self, sv: float, pv: float) -> float:
        err = sv - pv if not self.neg else pv - sv
        self.integral += err
        derivative = err - self.prev_error
        self.prev_error = err
        return self.p * err + self.i * self.integral + self.d * derivative


class BaseBetaPID(Callback):
    def __init__(self, target_ebops: float, p: float, i: float, d: float = 0.0) -> None:
        assert target_ebops > 0, 'Target EBOPs must be greater than 0.'
        self.pid = PID(p, i, d, neg=True)
        self.target_ebops = target_ebops

    def get_ebops(self):
        assert isinstance(self.model, Model)
        ebops: float = 0.0
        for layer in self.model.layers:
            if hasattr(layer, 'ebops'):
                ebops += float(layer.ebops)
        return ebops

    def set_beta(self, beta: float):
        assert isinstance(self.model, Model)
        for layer in self.model._flatten_layers():
            if getattr(layer, '_beta', None) is not None:
                layer._beta.assign(ops.convert_to_tensor(beta, dtype=layer._beta.dtype))

    def on_epoch_begin(self, epoch, logs: dict | None = None):
        assert isinstance(logs, dict)
        ebops = self.get_ebops()
        beta = self.pid(self.target_ebops, ebops)
        beta = max(beta, 0.0)
        self.set_beta(beta)
        logs['beta'] = beta
        logs['ebops'] = ebops


class BetaPID(BaseBetaPID):
    """
    Control the beta value of the Q Layers using a PID controller to reach a specified target EBOPs.

    Parameters
    ----------
    target_ebops : float
        The target EBOPs to reach.
    init_beta : float, optional
        The initial beta value to set before training starts. If None, the average beta of the model is used.
        If initial beta is set, it will be applied to the model at the beginning of training.
    p : float, default 1.0
        The proportional gain of the PID controller.
    i : float, default 2e-3
        The integral gain of the PID controller.
    d : float, default 0.0
        The derivative gain of the PID controller. As EBOPs is noisy, it is recommended to set this to 0.0 or a very small value.
    warmup : int, default 10
        The number of epochs to warm up the beta value. During this period, the beta value will not be updated.
    log : bool, default True
        If True, the beta value and error in EBOPs will be processed in logarithmic scale.
    max_beta : float, default float('inf')
        The maximum beta value to set. If the computed beta exceeds this value, it will be clamped to this maximum.
    min_beta : float, default 0.0
        The minimum beta value to set. If the computed beta is below this value, it will be clamped to this minimum.
    damp_beta_on_target : float, default 0.0
        A damping factor applied to the beta value when the target EBOPs is reached: beta *= (1 - damp_beta_on_target).
        This can help mitigating beta overshooting.
    """

    def __init__(
        self,
        target_ebops: float,
        init_beta: float | None = None,
        p: float = 1.0,
        i: float = 2e-3,
        d: float = 0.0,
        warmup: int = 10,
        log: bool = True,
        max_beta: float = float('inf'),
        min_beta: float = 0.0,
        damp_beta_on_target: float = 0.0,
    ) -> None:
        super().__init__(target_ebops, p, i, d)
        self.warmup = warmup
        self.init_beta = init_beta
        self.max_beta = max_beta
        self.min_beta = min_beta
        self.damp_beta_on_target = damp_beta_on_target
        self.log = log

    def get_avg_beta(self) -> float:
        assert isinstance(self.model, Model)
        n, c = 0, 0.0
        for layer in self.model._flatten_layers():
            if hasattr(layer, '_beta'):
                c += float(layer._beta.numpy())
                n += 1
        return c / n if n > 0 else 0.0

    def on_train_begin(self, logs=None):
        if self.init_beta is not None:
            self.set_beta(self.init_beta)
        self.beta = self.get_avg_beta() if self.init_beta is None else self.init_beta
        self._ebops = self.get_ebops()

    def on_epoch_begin(self, epoch: int, logs: dict | None = None) -> None:
        assert isinstance(logs, dict)

        if epoch < self.warmup:
            if self.init_beta is not None:
                self.set_beta(self.init_beta)
            return

        ebops = self._ebops
        if epoch == self.warmup:
            # match integral term ST init beta = beta for the first epoch
            # beta = p * P + i * I
            if not self.log:
                err = 1 - ebops / self.target_ebops
                self.pid.integral = (self.beta - self.pid.p * err) / self.pid.i - err
            else:
                err = log10(ebops / self.target_ebops + 1e-9)
                self.pid.integral = (log10(self.beta) - self.pid.p * err) / self.pid.i - err

        if self.log:
            beta = 10.0 ** self.pid(0, log10(ebops / self.target_ebops))
        else:
            beta = self.pid(1, ebops / self.target_ebops)

        if ebops < self.target_ebops:
            beta *= 1 - self.damp_beta_on_target

        beta = max(min(beta, self.max_beta), self.min_beta)
        self.set_beta(beta)
        self.beta = beta

    def on_epoch_end(self, epoch: int, logs: dict | None = None):
        assert isinstance(logs, dict)
        logs['beta'] = self.beta
        ebops = self.get_ebops()
        logs['ebops'] = ebops
        self._ebops = ebops
